"""
Tag Utils
"""

import logging

import sqlalchemy.ext.declarative
import sqlalchemy.orm

import wayround_org.utils.terminal
import wayround_org.utils.db


class TagEngine(wayround_org.utils.db.BasicDB):

    def __init__(
            self,
            config_string=None,
            bind=None,
            decl_base=None,
            metadata=None,
            init_table_data='tag'
            ):

        super().__init__(
            config_string=config_string,
            bind=bind,
            decl_base=decl_base,
            metadata=metadata,
            init_table_data=init_table_data
            )

        return

    def init_table_mappings(self, init_table_data):

        class Tag(self.decl_base):

            __tablename__ = init_table_data

            tid = sqlalchemy.Column(
                sqlalchemy.Integer,
                primary_key=True,
                autoincrement=True
                )

            obj = sqlalchemy.Column(
                sqlalchemy.UnicodeText,
                nullable=False,
                default='',
                index=True
                )

            tag = sqlalchemy.Column(
                sqlalchemy.UnicodeText,
                nullable=False,
                default='',
                index=True
                )

        self.Tag = Tag

        return

    def get_mapped_tag_table(self):
        ret = None
        if self.Tag.__tablename__ in self.decl_base.metadata.tables:
            ret = self.decl_base.metadata.tables[self.Tag.__tablename__]
        return ret

    def create_tables(self):
        self.get_mapped_tag_table().create(checkfirst=True)
        return

    def set_many_objects_tags(
            self,
            obj_tags_2tuple_list,
            commit_every=1000
            ):
        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        commit_counter = 0

        for obj, tags in obj_tags_2tuple_list:

            if tags is None:
                tags = []

            session.query(self.Tag).filter_by(obj=obj).delete()

            for i in tags:
                a = self.Tag()
                a.obj = obj
                a.tag = i
                session.add(a)

            commit_counter += 1

            if commit_counter >= commit_every:
                logging.debug("Committing")
                session.commit()
                commit_counter = 0

        session.commit()
        session.close()

        return

    def set_object_tags(self, obj, tags=None):
        return self.set_many_objects_tags(
            [(obj, tags)]
            )

    def get_object_tags(self, obj):

        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        q = session.query(self.Tag).filter_by(obj=obj).all()

        ret = set()
        for i in q:
            ret.add(i.tag)

        session.close()

        return list(ret)

    def get_objects(self):

        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        q = session\
            .query(sqlalchemy.distinct(self.Tag.obj))\
            .order_by(self.Tag.obj)\
            .all()

        ret = list()

        for i in q:
            ret.append(i[0])

        session.close()

        return ret

    def get_objects_and_tags_dict(self):

        ret = dict()

        objs = self.get_objects()

        for i in objs:
            ret[i] = self.get_tags(i)

        return ret

    def get_all_tags(self):
        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        q = session.query(sqlalchemy.distinct(self.Tag.tag)).all()

        ret = []
        for i in q:
            ret.append(i[0])

        session.close()

        return ret

    def get_size(self):
        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        ret = session.query(self.Tag).count()

        session.close()

        return ret

    def get_objects_by_tag(self, tag):
        return self.get_objects_by_tags([tag])

    def get_objects_by_tags(self, tags):
        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        ret = set()

        q = session.query(self.Tag)\
            .filter(self.Tag.tag.in_(tags))\
            .all()

        for i in q:
            ret.add(i.obj)

        session.close()

        return list(ret)

    def del_object_tags(self, obj, synchronize_session='evaluate'):

        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        if isinstance(obj, list):

            if len(obj) > 0:
                for i in range(int(len(obj) / 100) + 1):
                    session.query(self.Tag)\
                        .filter(self.Tag.obj.in_(
                            obj[i * 100:(i + 1) * 100]
                            )
                        )\
                        .delete(
                            synchronize_session=synchronize_session
                            )

        else:
            session.query(self.Tag).filter_by(obj=obj).delete(
                synchronize_session=synchronize_session
                )

        session.commit()
        session.close()

        return

    def del_objects_by_tags(self, tags):

        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        session.query(self.Tag)\
            .filter(self.Tag.tag.in_(tags))\
            .delete()

        session.commit()
        session.close()

        return

    def remove_duplicated_objects(self, mute=True):

        session = sqlalchemy.orm.Session(self.decl_base.metadata.bind)

        logging.info("Loading...")
        objs = self.get_objects()
        logging.info("Cleaning...")

        changed = False
        ii = 0
        ic = len(objs)
        removed = 0
        for i in objs[:]:

            if objs.count(i) > 1:

                changed = True

                while i in objs:
                    objs.remove(i)

                removed += \
                    session.query(self.Tag)\
                    .filter_by(obj=i)\
                    .count()

                session.query(self.Tag)\
                    .filter_by(obj=i)\
                    .delete()

            ii += 1

            if not mute:
                wayround_org.utils.terminal.progress_write(
                    "    {} of {} ({:.2f}%, deleted {})".format(
                        ii,
                        ic,
                        100 / (float(ic / ii)),
                        removed
                        )
                    )

        if not mute:
            wayround_org.utils.terminal.progress_write_finish()

        if changed:
            session.commit()

        session.close()

        return

    def clear(self):
        self.get_mapped_tag_table().delete()
        return




import logging
import os
import select
import socket
import ssl
import threading
import time

import wayround_org.utils.threading

CAT_READWRITE_TYPES = ['sync', 'async']

SELECT_SLEEP = 0.02
UNSELECTABLE_SLEEP = SELECT_SLEEP


class CatTerminationFlagFound(Exception):
    pass


class Streamer:

    def __init__(
            self,
            stream_object,
            stream_object_read_meth,
            stream_object_write_meth,
            stream_object_mode='sync',
            stream_object_selectable=True,
            stream_object_unselectable_sleep=0.2,
            bs=2 * 1024 ** 2,
            descriptor_to_wait_for=None,
            flush_after_each_write=False,
            standard_write_method_result=None,
            thread_name='Thread',
            termination_event=None,
            verbose=False,
            debug=False,
            on_exit_callback=None,
            on_input_read_error=None,
            on_output_write_error=None
            ):

        if not stream_object_mode in CAT_READWRITE_TYPES:
            raise ValueError("invalid `descriptor_mode'")

        if bs < 1:
            raise ValueError("invalid `bs'")

        if not standard_write_method_result in [None, False, True]:
            raise ValueError("invalid `standard_write_method_result'")

        if not isinstance(stream_object_selectable, bool):
            raise TypeError("`stream_object_selectable' must be bool")

        self._verbose = verbose
        self._debug = debug

        self._bs = bs

        self._stream_object = stream_object
        self._stream_object_mode = stream_object_mode
        self._stream_object_selectable = stream_object_selectable
        self._stream_object_unselectable_sleep = \
            stream_object_unselectable_sleep

        self._stream_object_read_meth = stream_object_read_meth
        self._stream_object_write_meth = stream_object_write_meth

        self._descriptor_to_wait_for = descriptor_to_wait_for

        self._standard_write_method_result = standard_write_method_result

        self._thread_name = thread_name

        self._flush_after_each_write = flush_after_each_write

        self._on_exit_callback = on_exit_callback
        self._on_input_read_error = on_input_read_error
        self._on_output_write_error = on_output_write_error

        self._termination_event = termination_event

        return

    def _wait_input_avail(self):

        if self._debug:
            logging.debug(
                "{}: waiting for input descriptor {}".format(
                    self._thread_name,
                    self._descriptor_to_wait_for
                    )
                )

        while len(
                select.select(
                    [self._descriptor_to_wait_for],
                    [],
                    [],
                    SELECT_SLEEP
                    )[0]
                ) == 0:

            if (self._termination_event
                    and self._termination_event.is_set()):
                raise CatTerminationFlagFound()

            if self._debug:
                logging.debug(
                    "{}: rewaiting for input descriptor {}".format(
                        self._thread_name,
                        self._descriptor_to_wait_for
                        )
                    )

            # the ssl.SSLSocket.pending() method is undocumented.
            # as so, I can't use it her, as it can be removed from ssl
            # socket class object at any time.
            # Instead of it I'll use time.sleep().
            # say thanks to /* Antoine Pitrou (pitrou) */
            # http://bugs.python.org/issue21430
            # .. commenting out

            # if isinstance(self._stream_object, ssl.SSLSocket):
            #     if self._debug:
            #         logging.debug(
            #             "{}: {} pending {}".format(
            #                 self._thread_name,
            #                 self._stream_object,
            #                 self._stream_object.pending()
            #                 )
            #             )
            #
            #     if self._stream_object.pending() != 0:
            #         if self._debug:
            #             logging.debug(
            #                 "{}: receiving pending data".format(
            #                     self._thread_name
            #                     )
            #                 )
            #         break

        if self._debug:
            logging.debug(
                "{}: input descriptor - ready - {}".format(
                    self._thread_name,
                    self._descriptor_to_wait_for
                    )
                )

        return

    def _wait_output_avail(self):

        if self._debug:
            logging.debug(
                "{}: waiting for output descriptor {}".format(
                    self._thread_name,
                    self._descriptor_to_wait_for
                    )
                )

        while len(
                select.select(
                    [],
                    [self._descriptor_to_wait_for],
                    [],
                    SELECT_SLEEP
                    )[1]
                ) == 0:

            if (self._termination_event
                    and self._termination_event.is_set()):
                raise CatTerminationFlagFound()

            if self._debug:
                logging.debug(
                    "{}: rewaiting for output descriptor {}".format(
                        self._thread_name,
                        self._descriptor_to_wait_for
                        )
                    )

        if self._debug:
            logging.debug(
                "{}: output descriptor - ready - {}".format(
                    self._thread_name,
                    self._descriptor_to_wait_for
                    )
                )

        return

    def read(self, data_size=None):
        """
        ret[0] == True - stream closed
        """

        ret_closed = False
        ret_buff = None

        try:

            if self._stream_object_mode == 'sync':

                if (self._termination_event
                        and self._termination_event.is_set()):
                    raise CatTerminationFlagFound()

                if self._debug:
                    logging.debug(
                        "Reading `{}' bytes using `{}'".format(
                            self._bs,
                            self._stream_object_read_meth
                            )
                        )

                res = self._stream_object_read_meth(self._bs)

                if self._debug:
                    logging.debug(
                        "Readed `{}' bytes using `{}'".format(
                            len(res),
                            self._stream_object_read_meth
                            )
                        )

                if len(res) == 0:
                    ret_closed = True
                    ret_buff = None
                    res = None

                else:

                    ret_closed = False
                    ret_buff = res
                    res = None

            elif self._stream_object_mode == 'async':

                if self._stream_object_selectable:
                    if self._descriptor_to_wait_for is not None:
                        self._wait_input_avail()
                else:
                    time.sleep(self._stream_object_unselectable_sleep)

                while True:

                    if (self._termination_event
                            and self._termination_event.is_set()):
                        raise CatTerminationFlagFound()

                    try:
                        if data_size is None:
                            ret_buff = self._stream_object_read_meth(self._bs)
                        else:
                            if data_size > self._bs:
                                ret_buff = self._stream_object_read_meth(
                                    self._bs
                                    )
                            else:
                                ret_buff = self._stream_object_read_meth(
                                    data_size
                                    )

                    except BlockingIOError:
                        pass

                    except ssl.SSLWantReadError:
                        # this is commented out, because withous use of
                        # .pending() too many garbage debugging messages
                        # are printed
                        # logging.debug("ssl.SSLWantReadError")
                        if self._stream_object_selectable:
                            select.select(
                                [self._descriptor_to_wait_for],
                                [],
                                [],
                                SELECT_SLEEP
                                )
                        else:
                            time.sleep(self._stream_object_unselectable_sleep)

                    except ssl.SSLWantWriteError:
                        # this is commented out, because withous use of
                        # .pending() too many garbage debugging messages
                        # are printed
                        # logging.debug("ssl.SSLWantWriteError")
                        if self._stream_object_selectable:
                            select.select(
                                [],
                                [self._descriptor_to_wait_for],
                                [],
                                SELECT_SLEEP
                                )
                        else:
                            time.sleep(self._stream_object_unselectable_sleep)
                    else:
                        break

            else:
                raise Exception(
                    "Programming error: self._stream_object_mode == {}".format(
                        self._stream_object_mode
                        )
                    )

        except CatTerminationFlagFound:
            logging.debug(
                "Caught termination flag in read method of {}:{}".format(
                    self,
                    self._thread_name
                    )
                )

            ret_closed = True
            ret_buff = None

            raise

        except:
            if self._on_input_read_error:
                threading.Thread(
                    target=self._on_input_read_error,
                    name="{} Input Read Error Thread".format(
                        self._thread_name
                        )
                    ).start()

            ret_closed = True
            ret_buff = None

            raise

        if not ret_closed and not isinstance(ret_buff, bytes):
            ret_closed = True
            raise TypeError(
                (
                    "Can read only bytes buffer "
                    "(Not str or anything other), but "
                    "buffer is ({}):{}..."
                    ).format(
                    type(ret_buff),
                    repr(ret_buff)[:100]
                    )
                )

        ret = ret_closed, ret_buff

        return ret

    def write(self, buff):

        if buff:

            if (self._termination_event
                    and self._termination_event.is_set()):
                raise CatTerminationFlagFound()

            try:

                while len(buff) != 0:

                    if self._stream_object_selectable:
                        if self._descriptor_to_wait_for is not None:
                            self._wait_output_avail()
                    else:
                        time.sleep(self._stream_object_unselectable_sleep)

                    sb = buff[:self._bs]
                    buff = buff[self._bs:]
                    self._stream_object_write_meth(sb)

                    if self._flush_after_each_write:
                        self._stream_object.flush()

            except TypeError as err_val:
                if err_val.args[0] == 'must be str, not bytes':
                    logging.warning(
                        "{}: hint: check that output is in bytes mode or do"
                        " conversion with convert_to_str option".format(
                            self._thread_name
                            )
                        )
                raise

            except CatTerminationFlagFound:
                logging.debug(
                    "Caught termination flag in write method of {}:{}".format(
                        self,
                        self._thread_name
                        )
                    )

                raise

            except BrokenPipeError:
                logging.exception(
                    "Broken pipe error arised. but I'll not give a damn\n"
                    " (take surraunding exit codes as a consideration of success)"
                    )
                # FIXME: cyrus-sasl-2.1.26.tar.gz
                #        (md5: a7f4e5e559a0e37b3ffc438c9456e425)
                #        gives an unknown broken pipe error.
                #        need to figure out this

                raise

            except:

                logging.error(
                    "{}: Can't use object's `{}' `{}' method.\n"
                    "    (possibly it isn't and error)".format(
                        self._thread_name,
                        self._stream_object,
                        self._stream_object_write_meth
                        )
                    )

                if self._on_output_write_error:
                    threading.Thread(
                        target=self._on_output_write_error,
                        name="`{}' Output Error Thread".format(
                            self._thread_name
                            )
                        ).start()

                raise

        return


def _is_cat_cycles_counter_enabled(verbose, count):
    return verbose or count is not None


def _is_cat_transferred_data_size_counter_enabled(
        verbose,
        maximum_total_data_transfer_size
        ):
    return verbose or maximum_total_data_transfer_size is not None


def cat(
        stdin,
        stdout,

        bs=2 * 1024 ** 2,
        count=None,
        maximum_total_data_transfer_size=None,

        threaded=False,
        thread_name='Thread',

        convert_to_str=None,

        read_method_name='read',
        write_method_name='write',

        read_type='sync',
        read_selectable=True,
        read_unselectable_sleep=UNSELECTABLE_SLEEP,

        write_type='sync',
        write_selectable=False,
        write_unselectable_sleep=UNSELECTABLE_SLEEP,

        exit_on_input_eof=True,
        flush_after_each_write=False,
        flush_on_input_eof=False,
        close_output_on_eof=False,

        descriptor_to_wait_for_input=None,
        descriptor_to_wait_for_output=None,

        apply_input_seek=True,
        apply_output_seek=True,

        standard_write_method_result=True,

        termination_event=None,

        verbose=False,
        debug=False,

        on_exit_callback=None,
        on_input_read_error=None,
        on_output_write_error=None
        ):

    if not read_method_name.isidentifier():
        raise ValueError("Wrong `read_method_name' parameter")

    if not write_method_name.isidentifier():
        raise ValueError("Wrong `write_method_name' parameter")

    if not read_type in CAT_READWRITE_TYPES:
        raise ValueError(
            "`read_type' must be in {}, not {}".format(
                CAT_READWRITE_TYPES,
                read_type
                )
            )

    if not write_type in CAT_READWRITE_TYPES:
        raise ValueError(
            "`write_type' must be in {}, not {}".format(
                CAT_READWRITE_TYPES,
                write_type
                )
            )

    if not hasattr(stdin, read_method_name):
        raise ValueError(
            "Object `{}' have no '{}' method".format(stdin, read_method_name)
            )

    if not hasattr(stdout, write_method_name):
        raise ValueError(
            "Object `{}' have no '{}' method".format(stdout, write_method_name)
            )

    if convert_to_str == True:
        convert_to_str = 'utf-8'
    elif convert_to_str == False:
        convert_to_str = None

    if (convert_to_str is not None
        and not isinstance(convert_to_str, str)
        ):
        raise ValueError(
            "convert_to_str can only be str(encoding name), bool or None"
            )

    if thread_name is None:
        thread_name = 'Thread'

    if threaded:

        return threading.Thread(
            target=cat,
            args=(stdin, stdout),
            kwargs=dict(
                bs=bs,
                count=count,
                threaded=False,
                thread_name=thread_name,
                verbose=verbose,
                convert_to_str=convert_to_str,
                read_method_name=read_method_name,
                write_method_name=write_method_name,
                read_type=read_type,
                read_selectable=read_selectable,
                read_unselectable_sleep=read_unselectable_sleep,
                write_type=write_type,
                write_selectable=write_selectable,
                write_unselectable_sleep=write_unselectable_sleep,
                exit_on_input_eof=exit_on_input_eof,
                flush_after_each_write=flush_after_each_write,
                flush_on_input_eof=flush_on_input_eof,
                close_output_on_eof=close_output_on_eof,
                descriptor_to_wait_for_input=descriptor_to_wait_for_input,
                descriptor_to_wait_for_output=descriptor_to_wait_for_output,
                apply_input_seek=apply_input_seek,
                apply_output_seek=apply_output_seek,
                standard_write_method_result=standard_write_method_result,
                termination_event=termination_event,
                on_exit_callback=on_exit_callback,
                on_input_read_error=on_input_read_error,
                on_output_write_error=on_output_write_error,
                debug=debug
                ),
            name=thread_name
            )

    #    if termination_event and termination_event.is_set():
    #        return None

    if verbose:
        logging.info("Starting `{}' thread".format(thread_name))

    buff = None

    if _is_cat_cycles_counter_enabled(verbose, count):
        c = 0

    if _is_cat_transferred_data_size_counter_enabled(
            verbose,
            maximum_total_data_transfer_size
            ):
        bytes_counter = 0

    if apply_input_seek and hasattr(stdin, 'seek'):
        try:
            stdin.seek(0)
        except:
            logging.exception("Can't seek")

    read_method = getattr(stdin, read_method_name)
    write_method = getattr(stdout, write_method_name)

    s1 = Streamer(
        stdin,
        read_method,
        None,
        stream_object_mode=read_type,
        stream_object_selectable=read_selectable,
        stream_object_unselectable_sleep=read_unselectable_sleep,
        bs=bs,
        descriptor_to_wait_for=descriptor_to_wait_for_input,
        thread_name='Input Streamer {}'.format(thread_name),
        termination_event=termination_event,
        verbose=verbose,
        debug=debug
        )

    s2 = Streamer(
        stdout,
        None,
        write_method,
        stream_object_mode=write_type,
        stream_object_selectable=write_selectable,
        stream_object_unselectable_sleep=write_unselectable_sleep,
        bs=bs,
        descriptor_to_wait_for=descriptor_to_wait_for_output,
        thread_name='Output Streamer {}'.format(thread_name),
        flush_after_each_write=flush_after_each_write,
        termination_event=termination_event,
        verbose=verbose,
        debug=debug
        )

    try:
        while True:

            if termination_event and termination_event.is_set():
                raise CatTerminationFlagFound()

            if maximum_total_data_transfer_size is None:
                closed, buff = s1.read()
            else:
                if bytes_counter + bs > maximum_total_data_transfer_size:
                    closed, buff = s1.read(
                        maximum_total_data_transfer_size - bytes_counter
                        )
                else:
                    closed, buff = s1.read()

            if not closed and buff is not None and len(buff) != 0:

                if debug:

                    buff_len = len(buff)
                    logging.debug(
                        "{}: Readed  {} bytes using method `{}'".format(
                            thread_name,
                            buff_len,
                            read_method
                            )
                        )
                    logging.debug(
                        "{}: buff data: {}".format(
                            thread_name,
                            repr(buff)
                            )
                        )
                    logging.debug(
                        "{}: Writing {} bytes using method `{}'".format(
                            thread_name,
                            buff_len,
                            write_method
                            )
                        )
                if convert_to_str is not None:
                    buff = str(buff, encoding=convert_to_str)

                if termination_event and termination_event.is_set():
                    raise CatTerminationFlagFound()

                s2.write(buff)

            else:
                if exit_on_input_eof:
                    break

            if _is_cat_cycles_counter_enabled(verbose, count):
                c += 1

            if isinstance(buff, (bytes, str)):
                if _is_cat_transferred_data_size_counter_enabled(
                        verbose,
                        maximum_total_data_transfer_size
                        ):

                    if isinstance(buff, bytes):
                        bytes_counter += len(buff)
                    if isinstance(buff, str):
                        bytes_counter += len(bytes(buff, 'utf-8'))

            if count is not None and count == c:
                break

            if (maximum_total_data_transfer_size is not None
                    and bytes_counter >= maximum_total_data_transfer_size):
                break

    except CatTerminationFlagFound:
        if debug:
            logging.debug(
                "{}: Termination flag caught".format(thread_name)
                )

    except:
        logging.exception(
            "{}: Exception in cat thread".format(thread_name)
            )

    if flush_on_input_eof:
        stdout.flush()

    if apply_output_seek and hasattr(stdout, 'seek'):
        try:
            stdout.seek(0, os.SEEK_END)
        except:
            pass

    if close_output_on_eof:
        if verbose:
            logging.info(" {}: Closing thread stdout".format(thread_name))

        stdout.close()

    if verbose:
        logging.info("""\
Ending `{name}' thread
    {{
       {num} cycles worked,
       {size} bytes ({sizem:4.2f} MiB) transferred,
       with buffer size {bufs} bytes ({bufm:4.2f} MiB)
    }}
""".format_map(
            {
                'name': thread_name,
                'num': c,
                'size': bytes_counter,
                'sizem': (float(bytes_counter) / 1024 / 1024),
                'bufs': bs,
                'bufm': (float(bs) / 1024 / 1024)
                }
            )
            )

    if on_exit_callback:
        threading.Thread(
            target=on_exit_callback,
            name="{} Exited Callback Thread".format(thread_name)
            ).start()

    return


def cat_socket_to_socket(
        one, another, bs=2 * 1024**2
        ):
    ret = cat(
        one,
        another,

        bs=bs,

        threaded=True,

        read_method_name='recv',
        write_method_name='send',

        read_type='async',
        read_selectable=False,
        read_unselectable_sleep=UNSELECTABLE_SLEEP,

        write_type='async',
        write_selectable=False,
        write_unselectable_sleep=UNSELECTABLE_SLEEP,

        exit_on_input_eof=False,
        flush_after_each_write=False,
        flush_on_input_eof=False,
        close_output_on_eof=False,

        descriptor_to_wait_for_input=one.fileno(),
        descriptor_to_wait_for_output=another.fileno(),

        apply_input_seek=False,
        apply_output_seek=False,

        standard_write_method_result=True,

        termination_event=None,

        verbose=False,
        debug=False,

        on_exit_callback=None,
        on_input_read_error=None,
        on_output_write_error=None
        )
    return ret


def lbl_write(
        stdin,
        stdout,
        threaded=False,
        typ='info'
        ):

    if not typ in ['info', 'error', 'warning']:
        raise ValueError("Wrong `typ' value")

    if threaded:
        return threading.Thread(
            target=lbl_write,
            args=(stdin, stdout),
            kwargs=dict(threaded=False)
            )
    else:

        while True:
            l = stdin.readline()
            if isinstance(l, bytes):
                l = l.decode('utf-8')
            if l == '':
                break
            else:
                l = l.rstrip(' \0\n')

                if typ == 'info':
                    stdout.info(l)
                elif typ == 'error':
                    stdout.error(l)
                elif typ == 'warning':
                    stdout.warning(l)

        return

    return


class SocketStreamer:

    """
    Featured class for flexibly handling socket connection

    Signals:
    'start' (self, self.socket)
    'stop' (self, self.socket)
    'error' (self, self.socket)
    'restart' (self, self.socket)
    'ssl wrap error' (self, self.socket)
    'ssl wrapped' (self, self.socket)
    'ssl ununwrapable' (self, self.socket)
    'ssl unwrap error' (self, self.socket)
    'ssl unwrapped' (self, self.socket)
    """

    def __init__(self, sock, socket_transfer_size=4096, debug=False):

        self.signal = wayround_org.utils.threading.Signal(
            self,
            ['start',
             'stop',
             'error',
             'restart',
             'ssl wrap error',
             'ssl wrapped',
             'ssl ununwrapable',
             'ssl unwrap error',
             'ssl unwrapped'
             ]
            )

        if sock.gettimeout() != 0:
            raise ValueError("`sock' timeout must be 0")

        if not isinstance(sock, (socket.socket, ssl.SSLSocket)):
            raise TypeError(
                "sock must be of type socket.socket or ssl.SSLSocket"
                )

        self.socket = sock
        self._socket_transfer_size = socket_transfer_size
        self._debug = debug

        self._pipe_inside = os.pipe2(os.O_NONBLOCK)
        self._pipe_outside = os.pipe2(os.O_NONBLOCK)

        # From remote process to current process.
        # For instance internals.
        self._strout = open(self._pipe_outside[1], 'wb', buffering=0)
        # For instance user.
        self.strout = open(self._pipe_outside[0], 'rb', buffering=0)

        # From current process to remote process.
        # For instance internals.
        self._strin = open(self._pipe_inside[0], 'rb', buffering=0)
        # For instance user.
        self.strin = open(self._pipe_inside[1], 'wb', buffering=0)

        # from strin to socket
        self._in_thread = None

        # from socket to strout
        self._out_thread = None

        self._connection_error_signalled = False
        self._connection_stop_signalled = False

        self._wrapping = False

        # flag indicating stop for instance threads
        self._stop_flag = False

        self._output_availability_watcher_thread = None
        self.connection = False

        self._in_thread_stop_event = threading.Event()
        self._out_thread_stop_event = threading.Event()

        return

    def __del__(self):
        try:
            self.destroy()
        except:
            logging.exception("Error destroying {}".format(self))
        return

    def destroy(self):
        self.stop()
        self._close_pipe_descriptors()
        return

    def get_socket(self):
        return self.socket

    def _close_pipe_descriptors(self):
        for i in [self._strout, self.strout, self._strin, self.strin]:
            if i:
                i.close()

        return

    def _start_threads(self):

        if self._stat_threads() == 'stopped':

            sock_selectable = True
            if self.is_ssl_working():
                sock_selectable = False

            self._in_thread = cat(
                stdin=self._strin,
                stdout=self.socket,
                threaded=True,
                write_method_name='send',
                close_output_on_eof=False,
                thread_name='strin -> socket',
                bs=self._socket_transfer_size,
                convert_to_str=None,
                read_method_name='read',
                read_type='async',
                read_selectable=True,
                write_type='async',
                write_selectable=sock_selectable,
                write_unselectable_sleep=UNSELECTABLE_SLEEP,
                exit_on_input_eof=True,
                descriptor_to_wait_for_input=self._strin,
                descriptor_to_wait_for_output=self.socket,
                apply_input_seek=False,
                apply_output_seek=False,
                flush_on_input_eof=False,
                on_exit_callback=self._on_in_thread_exit,
                on_output_write_error=self._on_socket_write_error,
                termination_event=self._in_thread_stop_event,
                flush_after_each_write=False,
                debug=self._debug
                )

            self._out_thread = cat(
                stdin=self.socket,
                stdout=self._strout,
                threaded=True,
                write_method_name='write',
                close_output_on_eof=False,
                thread_name='socket -> strout',
                bs=self._socket_transfer_size,
                convert_to_str=None,
                read_method_name='recv',
                read_type='async',
                read_selectable=sock_selectable,
                read_unselectable_sleep=UNSELECTABLE_SLEEP,
                write_type='async',
                write_selectable=True,
                exit_on_input_eof=True,
                descriptor_to_wait_for_input=self.socket,
                descriptor_to_wait_for_output=self._strout,
                apply_input_seek=False,
                apply_output_seek=False,
                flush_on_input_eof=True,
                on_exit_callback=self._on_out_thread_exit,
                on_input_read_error=self._on_socket_read_error,
                termination_event=self._out_thread_stop_event,
                flush_after_each_write=True,
                debug=self._debug
                )

            self._in_thread.start()
            self._out_thread.start()

            self._wait_threads('working')

        return

    def _stop_threads(self):

        if self._stat_threads() != 'stopped':

            t_in = self._in_thread
            t_out = self._out_thread

            self._in_thread_stop_event.set()
            self._out_thread_stop_event.set()

            if t_in is not None:
                t_in.join()

            if t_out is not None:
                t_out.join()

            self._wait_threads('stopped')

            self._in_thread_stop_event.clear()
            self._out_thread_stop_event.clear()

        return

    def _restart_threads(self):
        self._stop_threads()
        self._start_threads()

        self.signal.emit('restart', self, self.socket)

        return

    def _stat_threads(self):

        ret = 'unknown'

        v1 = self._in_thread
        v2 = self._out_thread

        if v1 is not None and v2 is not None:
            ret = 'working'

        elif v1 is None and v2 is None:
            ret = 'stopped'

        return ret

    def _send_connection_stopped_event(self):
        if not self._wrapping:
            if not self._connection_stop_signalled:
                self._connection_stop_signalled = True

                self.signal.emit('stop', self, self.socket)

        return

    def _send_connection_error_event(self):
        if not self._wrapping:
            if not self._connection_error_signalled:
                self._connection_error_signalled = True

                self.signal.emit('error', self, self.socket)

        return

    def start(self):

        if self.stat() == 'stopped':

            self._stop_flag = False

            self._connection_error_signalled = False
            self._connection_stop_signalled = False

            self._start_threads()

            self._output_availability_watcher_thread = threading.Thread(
                target=self._output_availability_watcher,
                name="Socket Output Availability Watcher Thread"
                )

            self._output_availability_watcher_thread.start()

            self.wait('working')

            self.signal.emit('start', self, self.socket)

        return

    def stop(self):

        if self.stat() != 'stopped':

            if self.is_ssl_working():
                self.stop_ssl()

            self._stop_flag = True

            self._stop_threads()

            self.wait('stopped')

            self.signal.emit('stop', self, self.socket)

        return

    def start_ssl(self, *args, **kwargs):
        """
        All parameters, same as for ssl.wrap_socket(). Exception is parameter
        socket, which
        taken from self.socket
        """

        if not self.is_ssl_working():

            self._wrapping = True

            if len(args) > 0:
                if issubclass(args[0], socket.socket):
                    del args[0]

            if 'sock' in kwargs:
                del kwargs['sock']

            kwargs['do_handshake_on_connect'] = False
            #            kwargs['suppress_ragged_eofs'] = False

            socket_wrap_result = None

            logging.debug("stopping threads before wrapping")

            self._stop_threads()

            logging.debug('before wrap sock is {}'.format(self.socket))

            try:
                socket_wrap_result = ssl.wrap_socket(
                    self.socket,
                    *args,
                    **kwargs
                    )

                while True:

                    try:
                        socket_wrap_result.do_handshake()

                    except ssl.SSLWantReadError:
                        select.select(
                            [socket_wrap_result.fileno()], [], [], SELECT_SLEEP
                            )

                    except ssl.SSLWantWriteError:
                        select.select(
                            [], [socket_wrap_result.fileno()], [], SELECT_SLEEP
                            )

                    except:
                        raise

                    else:
                        break

            except:
                logging.exception("ssl wrap error")
                self.signal.emit('ssl wrap error', self, self.socket)
            else:
                logging.debug(
                    """
peer cert:
{}
cipher:
{}
compression:
{}
""".format(
                        socket_wrap_result.getpeercert(binary_form=False),
                        socket_wrap_result.cipher(),
                        socket_wrap_result.compression()
                        )
                    )

                self.socket = socket_wrap_result

                #                self.socket.settimeout(0)

                logging.debug('after wrap sock is {}'.format(self.socket))

                logging.debug("starting threads after wrapping")

                self._start_threads()

                self.signal.emit('ssl wrapped', self, self.socket)

            self._wrapping = False

        return

    def stop_ssl(self):

        if self.is_ssl_working():

            if not self.connection:

                logging.debug(
                    "Connection already gone. "
                    "Unwrapping is pointless (and erroneous)"
                    )
                self.signal.emit('ssl ununwrapable', self, self.socket)

            else:

                self._wrapping = True

                logging.debug('before unwrap sock is {}'.format(self.socket))

                self._stop_threads()

                s = None
                try:
                    s = self.socket.unwrap()
                except:
                    logging.exception("ssl unwrap error")
                    self.signal.emit('ssl unwrap error', self, self.socket)

                else:
                    self.socket = s

                    self._start_threads()

                    logging.debug(
                        'after unwrap sock is {}'.format(self.socket)
                        )

                    self.signal.emit('ssl unwrapped', self, self.socket)

                self._wrapping = False

        return

    def _unwrap_procedure(self):
        if self.is_ssl_working():
            self.stop_ssl()
        else:
            logging.debug("Socket not wrapped - unwrapping not needed")
        return

    def is_ssl_working(self):
        return isinstance(self.socket, ssl.SSLSocket)

    def stat(self):

        ret = 'unknown'

        threads = self._stat_threads()
        v3 = self._output_availability_watcher_thread

        logging.debug(
            "{} :: status :: threads == {}, {}".format(self, threads, v3)
            )

        if threads == 'working':
            #  and v3 != None
            ret = 'working'

        elif threads == 'stopped' and v3 is None:
            ret = 'stopped'

        return ret

    def wait(self, what='stopped'):

        allowed_what = ['stopped', 'working']

        if not what in allowed_what:
            raise ValueError("`what' must be in {}".format(allowed_what))

        while True:
            logging.debug(
                "{} :: waiting for `{}`".format(
                    self.wait,
                    what
                    )
                )
            if self.stat() == what:
                break
            time.sleep(SELECT_SLEEP)

        return

    def _wait_threads(self, what='stopped'):

        allowed_what = ['stopped', 'working']

        if not what in allowed_what:
            raise ValueError("`what' must be in {}".format(allowed_what))

        while True:
            logging.debug(
                "{} :: waiting for `{}`".format(
                    self._wait_threads,
                    what
                    )
                )
            if self._stat_threads() == what:
                break
            time.sleep(SELECT_SLEEP)

        return

    def _on_in_thread_exit(self):
        self._in_thread = None
        self._any_thread_exited()
        return

    def _on_out_thread_exit(self):
        self._out_thread = None
        self._any_thread_exited()
        return

    def _any_thread_exited(self):

        if not self._wrapping:
            self.connection = False
            self._unwrap_procedure()
            self.socket.close()

        self._stop_threads()
        return

    def _on_socket_write_error(self):
        self._on_socket_read_write_error()
        return

    def _on_socket_read_error(self):
        self._on_socket_read_write_error()
        return

    def _on_socket_read_write_error(self):
        self.connection = False

        self.stop()

        self._send_connection_error_event()
        return

    def _output_availability_watcher(self):

        stopped_by_flag = False

        while len(
                select.select([], [self.socket.fileno()], [], SELECT_SLEEP)[1]
                ) == 0:

            if self._stop_flag:
                stopped_by_flag = True
                break

        if not stopped_by_flag:

            if not self._wrapping:

                self.connection = True
                self.signal.emit('start', self, self.socket)

        self._output_availability_watcher_thread = None

        return




import select
import ssl
import threading
import queue
import time

import wayround_org.mail.miscs

DEBUG_NB_FUNCS = False


def nb_handshake(sock, stop_event=None, select_timeout=0.5):

    while True:

        if stop_event is not None and stop_event.is_set():
            break

        try:
            data = sock.do_handshake()
        except BlockingIOError:
            select.select([sock], [], [], select_timeout)
        except ssl.SSLWantReadError:
            select.select([sock], [], [], select_timeout)
        except ssl.SSLWantWriteError:
            select.select([], [sock], [], select_timeout)
        except OSError as err:
            if err.errno == 9:
                break
            else:
                raise
        else:
            break

    return


def nb_recv(sock, bs=4096, stop_event=None, select_timeout=0.5):
    """

    if len() of returned value is 0, then foreign socket is closed

    not returns until something recived from socket or socket is
    closed. use stop_event to terminate this function

    stop_event must be of threading.Event() type
    """

    if stop_event is not None:
        if not isinstance(stop_event, threading.Event):
            raise TypeError("`stop_event' must be of threading.Event type")

    data = b''

    while True:

        if stop_event is not None and stop_event.is_set():
            break

        try:
            data = sock.recv(bs)
        except BlockingIOError:
            select.select([sock], [], [], select_timeout)
        except ssl.SSLWantReadError:
            select.select([sock], [], [], select_timeout)
        except ssl.SSLWantWriteError:
            select.select([], [sock], [], select_timeout)
        except OSError as err:
            if err.errno == 9:
                ret = b''
                break
            else:
                raise
        else:
            break

    if DEBUG_NB_FUNCS:
        print("recvd: {}".format(data))

    return data


def nb_sendall(sock, data, bs=4096, stop_event=None, select_timeout=0.5):

    if DEBUG_NB_FUNCS:
        print("sending: {}".format(data))

    if stop_event is not None:
        if not isinstance(stop_event, threading.Event):
            raise TypeError("`stop_event' must be of threading.Event type")

    if type(data) != bytes:
        raise TypeError("`data' must be bytes")

    while True:

        try:
            sock.sendall(data)
        except BlockingIOError:
            select.select([], [sock], [], select_timeout)
        except ssl.SSLWantReadError:
            select.select([sock], [], [], select_timeout)
        except ssl.SSLWantWriteError:
            select.select([], [sock], [], select_timeout)
        except OSError as err:
            if err.errno == 9:
                ret = b''
                break
            else:
                raise
        else:
            break

        if stop_event is not None and stop_event.is_set():
            break

    return


class LblRecvReaderBuffer:

    def __init__(
            self,
            sock,
            recv_size=4096,
            line_terminator=wayround_org.mail.miscs.STANDARD_LINE_TERMINATOR,
            maximum_line_length=wayround_org.mail.miscs.MAX_LINE_LENGTH
            ):
        """

        line_terminator=b'\r\n' (0x10, 0x13) is mail system message terminator
        style

        if maximum_line_length is None, then control of maximum line length is
        disabled.

        if maximum_line_length is not None, it must be positive int

        if maximum_line_length is set, then if working thread
        (_02_worker_thread) detects limitation reach - then this class
        instance property .max_line_error is set to True and .stop() method is
        called.
        """

        if maximum_line_length is not None:
            if not isinstance(maximum_line_length, int):
                raise TypeError("`maximum_line_length' must be None or int")

            if not maximum_line_length >= 1:
                raise ValueError("`maximum_line_length' must be >= 1")

        self.sock = sock
        self.line_terminator = line_terminator
        self.recv_size = recv_size
        self.maximum_line_length = maximum_line_length

        self._socket_is_closed = False
        self._line_terminator_len = len(self.line_terminator)

        self._01_worker_thread = None
        self._02_worker_thread = None

        self._stop_flag = threading.Event()

        self._recv_buffer_queue = queue.Queue()

        # this buffer is here and not in thread look function for make
        # available it's size information
        self._input_bytes_buff = b''
        self._input_bytes_buff_lock = threading.Lock()

        # this list not supposed to be accessible for outsiders
        # and the only way to get items from it is special method
        self._lines = []
        self._lines_lock = threading.Lock()

        self.max_line_error = False

        return

    def start(self):
        if self._01_worker_thread is None:

            self._01_worker_thread = threading.Thread(
                target=self._worker_thread_target_01
                )
            self._01_worker_thread.start()

        if self._02_worker_thread is None:

            self._02_worker_thread = threading.Thread(
                target=self._worker_thread_target_02
                )
            self._02_worker_thread.start()

        return

    def stop(self):
        self._stop_flag.set()
        self.wait()
        self._01_worker_thread = None
        self._02_worker_thread = None
        return

    def wait(self):
        if self._01_worker_thread is not None:
            self._01_worker_thread.join()

        if self._02_worker_thread is not None:
            self._02_worker_thread.join()

        return

    def _worker_thread_target_01(self):
        """
        Recive data from socket and put it in queue.

        with this thread I just want server to read input data fast as I
        think it is important.
        """

        while True:
            if self._stop_flag.is_set():
                break

            res = nb_recv(self.sock, stop_event=self._stop_flag)

            if len(res) == 0:
                self._socket_is_closed = True
                break

            self._recv_buffer_queue.put(res)

        threading.Thread(target=self.stop).start()

        return

    def _worker_thread_target_02(self):
        """
        Takes data from queue and glues it with bytes string, after
        what - searches and slices it by line separators, adding result to
        dedicated list. warning: line separator is not stripped from lines!
        """

        while True:

            if self._stop_flag.is_set():
                break

            res_empty = False
            try:
                res = self._recv_buffer_queue.get(block=True, timeout=0.5)
            except queue.Empty:
                res_empty = True

            if res_empty:
                continue

            with self._input_bytes_buff_lock:

                self._input_bytes_buff += res

                while True:
                    if self._stop_flag.is_set():
                        break

                    sep_pos = self._input_bytes_buff.find(
                        self.line_terminator
                        )

                    if sep_pos == -1:
                        break

                    cut_pos = sep_pos + self._line_terminator_len

                    res_line_bytes = self._input_bytes_buff[:cut_pos]

                    self._input_bytes_buff = self._input_bytes_buff[cut_pos:]

                    with self._lines_lock:

                        self._check_line_length_rtn(len(res_line_bytes))
                        if self.max_line_error:
                            break

                        self._lines.append(res_line_bytes)

                self._check_line_length_rtn(len(self._input_bytes_buff))

                if self.max_line_error:
                    break

        threading.Thread(target=self.stop).start()

        return

    def _check_line_length_rtn(self, length):
        if self.maximum_line_length is not None:
            if length >= self.maximum_line_length:
                self.max_line_error = True
        return

    def get_lines_bytes_size(self, stop_event=None):
        sum_ = 0
        with self._lines_lock:
            for i in self._lines:

                if stop_event is not None and stop_event.is_set():
                    break

                if (self._stop_flag is not None
                        and self._stop_flag.is_set()):
                    break

                sum_ += len(i)

        return sum_

    def get_next_bytes(
            self,
            size,
            delete_readen_data=True,
            stop_event=None
            ):
        """
        if delete_readen_data is True, then in case on successful read,
            readen lines and readen part of last readen line will be removed.
        """

        if size < 0:
            raise ValueError("`size' must be not < 0")

        ret = None
        if self.get_lines_bytes_size(stop_event) >= size:
            with self._lines_lock:
                buff = b''
                line_i = 0
                remained_size = size
                error_exit = False
                len_lines = len(self._lines)
                while True:

                    if stop_event is not None and stop_event.is_set():
                        error_exit = True
                        break

                    if (self._stop_flag is not None
                            and self._stop_flag.is_set()):
                        error_exit = True
                        break

                    if remained_size < 0:
                        error_exit = True
                        raise Exception("programming error")

                    if remained_size == 0:
                        break

                    if len_lines == line_i:
                        break

                    if len(self._lines[line_i]) > remained_size:
                        buff += self._lines[line_i][:remained_size]
                        remained_size = 0
                        break
                    else:
                        buff += self._lines[line_i]
                        remained_size -= len(self._lines[line_i])

                    line_i += 1

                if not error_exit:
                    ret = buff

                    if delete_readen_data:

                        del buff
                        del line_i
                        del remained_size
                        del error_exit
                        del len_lines

                        left_delete_bytes = size

                        while True:

                            if left_delete_bytes == 0:
                                break

                            len_lines0 = len(self._lines[0])

                            if len_lines0 > left_delete_bytes:
                                self._lines[0] = \
                                    self._lines[0][left_delete_bytes:]
                                left_delete_bytes = 0
                            else:
                                del self._lines[0]
                                left_delete_bytes -= len_lines0

        return ret

    def nb_get_next_bytes(
            self,
            size,
            delete_readen_data=True,
            stop_event=None,
            retry_interval=0.2
            ):
        """

        Uses get_next_bytes(): wait's until it returns something.
        Can be interupted using passed event object.
        """

        if not isinstance(stop_event, threading.Event):
            raise TypeError("`stop_event' must be of threading.Event type")

        ret = None

        while True:

            if stop_event is not None and stop_event.is_set():
                break

            if (self._stop_flag is not None
                    and self._stop_flag.is_set()):
                break

            ret = self.get_next_bytes(
                size,
                delete_readen_data=delete_readen_data,
                stop_event=stop_event
                )

            if ret is not None:
                break

            time.sleep(retry_interval)

        return ret

    def get_next_line(self):
        """
        Returns next line, or None if there is not yet any. If socket is
        closed - and empty line with terminator is returned.

        this method is async. if self._lines has any items, this method
        returns first of them, and removes it from self._lines.

        this method supposed to be the only way to get data from self._lines,
        as there shuld be thread safety to work with self._lines.
        """

        ret = None

        if self._socket_is_closed:
            ret = self.line_terminator
        else:
            with self._lines_lock:
                if len(self._lines) != 0:
                    ret = self._lines[0]
                    del self._lines[0]

        return ret

    def nb_get_next_line(self, stop_event, retry_interval=0.2):
        """

        Uses get_next_line(): wait's until it returns something.
        Can be interupted using passed event object.
        """

        if not isinstance(stop_event, threading.Event):
            raise TypeError("`stop_event' must be of threading.Event type")

        ret = None

        while True:

            if stop_event is not None and stop_event.is_set():
                break

            if (self._stop_flag is not None
                    and self._stop_flag.is_set()):
                break

            ret = self.get_next_line()

            if ret is not None:
                break

            time.sleep(retry_interval)

        return ret

    def get_bytes_buffer_byte_size(self):
        with self._input_bytes_buff_lock:
            ret = len(self._input_bytes_buff)
        return ret

    def get_lines_buffer_byte_size(self):
        ret = 0
        with self._lines_lock:
            for i in self._lines:

                if (self._stop_flag is not None
                        and self._stop_flag.is_set()):
                    break

                ret += len(i)
        return ret

    def get_lines_count(self):
        with self._lines_lock:
            ret = len(self._lines)
        return ret




import weakref
import logging
import threading

import wayround_org.utils.types


class WeakMethod:

    def __init__(self, method, callback=None):

        if not callable(method):
            raise ValueError("callable must be provided")

        if not wayround_org.utils.types.is_method(method):
            raise ValueError("provided callable not a method")

        if callback and not callable(callback):
            raise ValueError("callback must be callable")

        self._debug = False

        self._object = weakref.ref(method.__self__, self._object_finalizes)

        if self._debug:
            logging.debug(
                "{} installed weakref: {}".format(self, self._object)
                )

        self._callback = callback
        self._method_name = method.__name__

        return

    def _object_finalizes(self, wr):

        if self._debug:
            logging.debug("{} finalizing method: {}".format(self, wr))

        self._object = None
        self._method_name = None

        if self._callback:
            threading.Thread(
                target=self._callback,
                args=(self,)
                ).start()

    def __call__(self):

        ret = None

        if self._object:
            ret = getattr(self._object(), self._method_name, None)

        return ret




import hashlib
import logging
import os
import re

import wayround_org.utils.path
import wayround_org.utils.stream
import wayround_org.utils.terminal


def make_dir_checksums(
        dirname,
        output_filename,
        rel_to=None,
        conv_to_rooted=True,
        exclude=None,
        verbose=True
        ):

    ret = 0

    dirname = wayround_org.utils.path.abspath(dirname)

    if not os.path.isdir(dirname):
        logging.error("Not is dir {}".format(dirname))
        ret = 1

    else:

        try:
            sums_fd = open(output_filename, 'w')
        except:
            logging.exception("Error opening output file")
            ret = 2
        else:
            try:
                ret = make_dir_checksums_fo(
                    dirname,
                    sums_fd,
                    rel_to,
                    conv_to_rooted,
                    exclude,
                    verbose=verbose
                    )
            except:
                logging.exception("Error")
                ret = 2
            finally:
                sums_fd.close()

    return ret


def make_dir_checksums_fo(
        dirname,
        output_fileobj,
        rel_to=None,
        conv_to_rooted=True,
        exclude=None,
        verbose=True
        ):

    if exclude is not None and not isinstance(exclude, list):
        raise ValueError("`exclude' must be list or None")

    if not isinstance(rel_to, str):
        rel_to = dirname

    ret = 0

    dirname = wayround_org.utils.path.abspath(dirname)

    if rel_to is None:
        rel_to = dirname

    rel_to = wayround_org.utils.path.abspath(rel_to)

    if not os.path.isdir(dirname):
        logging.error("Not a dir {}".format(dirname))
        ret = 1

    else:

        if not hasattr(output_fileobj, 'write'):
            logging.error("Wrong output file object")
            ret = 2
        else:

            for wres in os.walk(dirname):

                wres[1].sort()
                wres[2].sort()

                root = wres[0]
                files = wres[2]

                for f in files:
                    root_f = wayround_org.utils.path.join(root, f)

                    if exclude is not None and root_f in exclude:
                        continue

                    rel_path = wayround_org.utils.path.relpath(
                        root_f, rel_to
                        )

                    if verbose:
                        wayround_org.utils.terminal.progress_write(
                            "    {}".format(rel_path)
                            )
                    if (os.path.isfile(root_f)
                        and
                        not os.path.islink(root_f)
                        ):
                        m = hashlib.sha512()
                        fd = None
                        try:
                            fd = open(root_f, 'rb')
                        except:
                            logging.exception(
                                "Can't open file `{}'".format(
                                    root_f
                                    )
                                )
                            ret = 3
                        else:
                            try:
                                while True:

                                    buf = fd.read()
                                    if len(buf) == 0:
                                        break

                                    m.update(buf)

                                wfn = rel_path

                                if (conv_to_rooted
                                        and not wfn.startswith(os.path.sep)):

                                    wfn = os.path.sep + wfn

                                output_fileobj.write(
                                    "{digest} *{pkg_file_name}\n".format_map(
                                        {
                                            'digest': m.hexdigest(),
                                            'pkg_file_name': wfn
                                            }
                                        )
                                    )
                            except:
                                logging.exception("Some error")
                            finally:
                                fd.close()

                        del(m)

    if verbose:
        wayround_org.utils.terminal.progress_write_finish()
    return ret


def is_dir_changed(dirname, checksum_file, method='sha512', verbose=False):

    ret = False

    checksum_file_tmp = checksum_file + '.tmp'

    wayround_org.utils.checksum.make_dir_checksums(
        dirname,
        checksum_file_tmp,
        rel_to='/',
        exclude=[
            checksum_file,
            checksum_file_tmp
            ],
        verbose=verbose
        )

    if not os.path.isfile(checksum_file):
        ret = True

    else:

        summ1 = wayround_org.utils.checksum.make_file_checksum(
            checksum_file
            )
        summ2 = wayround_org.utils.checksum.make_file_checksum(
            checksum_file_tmp
            )

        ret = summ1 != summ2

        os.unlink(checksum_file)

    os.rename(checksum_file_tmp, checksum_file)

    return ret


def make_file_checksum(filename, method='sha512'):
    ret = 0

    if not method.isidentifier() or not hasattr(hashlib, method):
        raise ValueError("hashlib doesn't have `{}'".format(method))

    try:
        f = open(filename, 'rb')
    except:
        logging.exception("Can't open file `{}'".format(filename))
        ret = 1
    else:
        summ = make_fileobj_checksum(f, method)
        if not isinstance(summ, str):
            logging.error("Can't get checksum for file `{}'".format(filename))
            ret = 2
        else:
            ret = summ

        f.close()

    return ret


def make_fileobj_checksum(fileobj, method='sha512'):
    ret = None
    hash_method_name = None

    if not method.isidentifier() or not hasattr(hashlib, method):
        raise ValueError("hashlib doesn't have `{}'".format(method))

    try:
        hash_method_name = eval("hashlib.{}()".format(method))
    except:
        logging.exception(
            "Error calling for hashlib method `{}'".format(method)
            )
        ret = 1
    else:
        wayround_org.utils.stream.cat(
            fileobj,
            hash_method_name,
            write_method_name='update',
            standard_write_method_result=False
            )
        ret = hash_method_name.hexdigest()
        del(hash_method_name)
    return ret


def parse_checksums_file_text(filename):
    ret = 0
    try:
        f = open(filename, 'rb')
    except:
        logging.exception("Can't open file `{}'".format(filename))
        ret = 1
    else:
        txt = f.read()
        f.close()
        sums = parse_checksums_text(txt)
        if not isinstance(sums, dict):
            logging.error(
                "Can't get checksums from file `{}'".format(filename)
                )
            ret = 2
        else:
            ret = sums

    return ret


def parse_checksums_text(text):
    ret = 0
    if isinstance(text, bytes):
        text = text.decode('utf-8')

    lines = text.splitlines()
    sums = {}
    for i in lines:
        ist = i.strip(' \n\t\0')
        if ist != '':
            re_res = re.match(r'(.*?) \*(.*)', ist)

            if re_res is None:
                ret = 1
                break
            else:
                sums[re_res.group(2)] = re_res.group(1)

    if ret == 0:
        ret = sums

    return ret


def checksums_by_list(file_lst, method):

    if not method.isidentifier() or not hasattr(hashlib, method):
        raise ValueError("hashlib doesn't have `{}'".format(method))

    ret = {}

    for i in file_lst:
        ret[i] = make_file_checksum(i, method=method)

    return ret


def render_checksum_dict_to_txt(sums_dict, sort=False):

    keys = list(sums_dict.keys())

    if sort:
        keys.sort()

    ret = ''

    for i in keys:
        ret += '{summ} *{path}\n'.format(summ=str(sums_dict[i]), path=str(i))

    return ret


def is_data_error(method, value, data):

    ret = True

    value = value.lower()

    if not method.isidentifier() or not hasattr(hashlib, method):
        raise ValueError("hashlib doesn't have `{}'".format(method))

    hash_method_name = eval("hashlib.{}()".format(method))

    hash_method_name.update(data)

    hd = hash_method_name.hexdigest().lower()

    return hd == value




import fcntl
import struct
import sys
import termios

import wayround_org.utils.text


def get_terminal_size(fd=1):
    res = None
    io_res = None
    arg = struct.pack('HHHH', 0, 0, 0, 0)

    try:
        io_res = fcntl.ioctl(
            fd,
            termios.TIOCGWINSZ,
            arg
            )
    except:
        res = None
    else:
        try:
            res = struct.unpack('HHHH', io_res)
        except:
            res = None

    if res is not None:
        res = {
            'ws_row': res[0],
            'ws_col': res[1],
            'ws_xpixel': res[2],
            'ws_ypixel': res[3]
            }

    return res


# TODO: output descriptor selector
def progress_write_finish():
    sys.stdout.write('\n')
    sys.stdout.flush()
    return


# TODO: output descriptor selector
def progress_write(line_to_write, new_line=False):

    new_line_str = ''

    if line_to_write.endswith('\n'):
        new_line = True
        line_to_write = line_to_write.rstrip('\n')

    if new_line:
        new_line = True
        new_line_str = '\n'

    width = 80
    ts = get_terminal_size(sys.stdout.fileno())
    if ts is not None:
        width = ts['ws_col']

    line_to_write_l = len(line_to_write)

    line_to_out = '\r{ltw}{spaces}{new_line}\r'.format_map(
        {
            'ltw': line_to_write,
            'spaces': ' ' * (width - line_to_write_l),
            'new_line': new_line_str
            }
        )

    if len(line_to_out) > width:
        line_to_out = line_to_out[:width + 1] + new_line_str + '\r'

    # TODO: put sys.stdout to parameters

    sys.stdout.write(line_to_out)
    sys.stdout.flush()
    return




import re
import logging

NEW_DICT_KEY_TEMPLATE = '{last_placer_child_name}_auto_appended_key_{num:010d}'
LAST_DICT_KEY_CHECK_RE = r'(.*)_auto_appended_key_(\d{10})'


def append(indict, what_to_append):

    last_name = max(indict)

    re_val = re.match(LAST_DICT_KEY_CHECK_RE, last_name)

    placement_i = 0

    if re_val:
        placement_i = int(re_val.group(2)) + 1
        new_key = NEW_DICT_KEY_TEMPLATE.format_map(
            {
                'last_placer_child_name': re_val.group(1),
                'num': placement_i
                }
            )
    else:
        new_key = NEW_DICT_KEY_TEMPLATE.format_map(
            {
                'last_placer_child_name': last_name,
                'num': placement_i
                }
            )

    logging.debug("Adding dict key: `{}'".format(new_key))
    indict[new_key] = what_to_append

    return




import bottle


class WSGIRefServer(bottle.ServerAdapter):

    def run(self, handler):

        import wsgiref.simple_server

        if self.quiet:

            class QuietHandler(wsgiref.simple_server.WSGIRequestHandler):

                def log_request(self, *args, **kw):
                    pass

            self.options['handler_class'] = QuietHandler

        self.srv = wsgiref.simple_server.make_server(
            self.host,
            self.port,
            handler,
            **self.options
            )

        self.srv.serve_forever()

        return




"""
Tools for filtering lists
"""

# TODO: future planed work
#       The aim if to create flexiplt and extensible filter function
#       to replace such in list.py and tarball.py


def filter_text_parse(filter_text):
    """
    Returns list of command structures

    ret = [
        dict(
            action   = '-' or '+',
            subject  = in ['path', 'filename', 'version', 'status'],
            function = <depends on subject> (no spaces allowed),
            data     = <depends on subject> (can contain spaces)
            )
        ]

    """
    ret = []

    lines = filter_text.splitlines()

    for i in lines:
        if i != '' and not i.isspace() and not i.startswith('#'):
            struct = i.split(' ', maxsplit=3)
            if not len(struct) == 4:
                logging.error("Wrong filter line: `{}'".format(i))
            else:
                struct = dict(
                    action=struct[0],
                    subject=struct[1],
                    function=struct[2],
                    data=struct[3]
                    )
                ret.append(struct)

    return ret


def filter(
        filter_text,
        full_source_list,
        subject_func_dict,
        compare_func_dict
        ):
    """
    parameters:
        filter_text,
            filter text str possibly supplyed by user

        full_source_list
            list with all possible variants

        subject_func_dict
            dict with subject functions
                keys   - are names of functions. any non-empty string without
                    spaces

                values - callables accepting following parameters:
                    1. this filter line subject func name
                        str with this function name
                    2. this filter line compare func name
                        which will be used
                    3. this one parameter is full item string which
                        is about to be checked.

                    callable must return some string value
                    which will be pased to comparison function
                    to be compared with it's parameter(s).
                    or return False to indicate what line already
                    didn't passed check on this stage, in which
                    case line being checked will be considered as absent in
                    full_source_list and will be minused

        compare_func_dict
            dict with comparisong functions
                keys   - are names of functions
                values - callables accepting following parameters:

                    1. this filter line compare func name
                        which will be used
                    2. this filter line subject func name
                        str with function name which was used for subject
                        devision
                    3. logical not: if true - will invert result
                    4. non case sensetivity - by default all functions
                        considered sensitive to case. this callabel can ignore
                        this parameter value if subject, for instance,
                        a numbers
                    5. filter parameter here goes 4-th value of of filter list 
                        line

    result: list with all, none or part of full_source_list items
    """

    parsed_filter_text = filter_text_parse(filter_text)
    
    

    return





def get_range_indexes(value, ranges_list):
    ret = []
    for r in ranges_list:
        if value in r:
            ret.append(ranges_list.index(r))
    return ret


def get_range_first_index(value, ranges_list):
    ret = -1
    for r in ranges_list:
        if value in r:
            ret = ranges_list.index(r)
            break
    return ret


def is_in_ranges(value, ranges_list):
    return get_range_first_index(value, ranges_list) != -1




def split(str_line):

    params = []

    str_line = str_line.strip()
    str_line_len = len(str_line)

    i = 0
    cur = 0

    if str_line_len > 0:
        while True:

            if i == str_line_len:
                params += work_param(str_line[cur:i + 1])
                break
            elif str_line[i] == ' ':
                if i > 0 and str_line[i - 1] == '\\':
                    i += 1
                    continue

                params += work_param(str_line[cur:i])
                cur = i
                while True:
                    if i == str_line_len:
                        break
                    if str_line[i] != ' ':
                        cur = i
                        break
                    i += 1
                continue
            elif str_line[i] in ['"', "'"]:
                if i > 0 and str_line[i - 1] == '\\':
                    i += 1
                    continue
                qc = str_line[i]
                i += 1

                if i == str_line_len:
                    continue

                p = find_quot(str_line, qc, i)
                if p is None:
                    i = str_line_len
                    continue
                i = p

                if i == str_line_len:
                    continue

                i += 1
                continue
            else:
                i += 1

    return params


def work_param(str_line):
    ret = correct_quotes_remove(str_line)
    ret = unquote_chars(ret)
    return [ret]


def correct_quotes_remove(str_line):
    ret = str_line

    q0 = None
    q1 = None
    cut0 = []
    cut1 = []
    cut2 = []

    for c in ['"', "'"]:
        while True:
            ret_len = len(ret)

            cut0 = [0]
            q0 = find_chars(ret, start=0, chars=c)

            if q0[0] is None:
                break

            cut0.append(q0[1])

            cut1 = [cut0[1] + 1]

            q1 = find_chars(ret, start=cut1[0], chars=c)

            if q1[0] is None:
                break

            cut1.append(q1[1])

            cut2 = [q1[1] + 1, ret_len]

            ret = ret[cut0[0]:cut0[1]] + \
                ret[cut1[0]:cut1[1]] + ret[cut2[0]:cut2[1]]

    return ret


def unquote_chars(str_line, chars=" '\""):
    clst = []
    for c in chars:
        clst.append(c)
    clst = sorted(set(clst))

    for c in clst:
        str_line.replace('\\' + c, c)

    return str_line


def find_chars(str_line, start=0, chars=" '\""):

    clst = []
    for c in chars:
        clst.append(c)
    clst = sorted(set(clst))
    lst = []

    for c in clst:
        cp = str_line.find(c, start)
        if cp != -1:
            if cp == 0 or (str_line[cp - 1] != '\\'):
                lst.append([c, cp])

    ret = None
    l_lst = len(lst)

    if l_lst == 0:
        ret = (None, None)
    elif l_lst == 1:
        ret = tuple(lst[0])
    else:
        ret = tuple(lst[0])

        for e in lst:
            if e[1] < ret[1]:
                ret = tuple(e)

    return ret


def find_quot(str_line, qc, start):

    ret = None

    cp = str_line.find(qc, start)
    if cp != -1:
        if cp == 0 or (str_line[cp - 1] != '\\'):
            ret = cp

    return ret




import os
import re

import wayround_org.utils.terminal
import wayround_org.utils.types
import wayround_org.utils.range


def columned_list_print(
        lst, width=None, columns=None,
        margin_right=' | ', margin_left=' | ', spacing=' | ',
        fd=1
        ):
    print(
        return_columned_list(
            lst, width=width, columns=columns,
            margin_right=margin_right, margin_left=margin_left,
            spacing=spacing, fd=fd
            )
        )


def return_columned_list(
        lst, width=None, columns=None,
        margin_right=' | ', margin_left=' | ', spacing=' | ',
        fd=1
        ):

    if not wayround_org.utils.types.struct_check(
            lst,
            {'t': list, '.': {'t': str}}
            ):
        raise TypeError("`lst' must be list of str")

    if width is None:
        if (
                (isinstance(fd, int) and os.isatty(fd))
                or (hasattr(fd, 'isatty') and fd.isatty())
                ):

            size = wayround_org.utils.terminal.get_terminal_size(fd)
            if size is None:
                width = 80
            else:
                width = size['ws_col']
        else:
            width = 80

    # print "width " + str(width)

    longest = 0
    lst_l = len(lst)
    for i in lst:
        l = len(i)
        if l > longest:
            longest = l

    mrr_l = len(margin_right)
    mrl_l = len(margin_left)
    spc_l = len(spacing)

    int_l = width - mrr_l - mrl_l

    if columns is None:
        columns = int((int_l / (longest + spc_l)))

    if columns < 1:
        columns = 1

    ret = ''
    for i in range(0, lst_l, columns):
        # print "i == " + str(i)
        l2 = lst[i:i + columns]

        l3 = []
        for j in l2:
            l3.append(j.ljust(longest))

        while len(l3) != columns:
            l3.append(''.ljust(longest))

        ret += "{mrl}{row}{mrr}\n".format_map(
            {
                'mrl': margin_left,
                'mrr': margin_right,
                'row': spacing.join(l3)
                }
            )

    return ret


def fill(char=' ', count=80):
    raise Exception("Deprecated")
    char = str(char)

    if len(char) < 1:
        char = ' '

    ret = char[0] * count
    return ret


def slice_string_to_sections(stri):
    return re.findall(r'[a-zA-Z]+|\d+|[\.\-\_\~\+]', stri)


def get_line_ranges(txt, nl='\n'):

    # TODO: make fix to correct lenght of nl

    ret = []

    x = 0

    while True:

        y = txt.find(nl, x)

        if y == -1:
            ret.append(range(x, len(txt)))
            break
        else:
            ret.append(range(x, y + 1))
            x = y + 1

    return ret


def get_line_index_at_offset(offset, ranges):
    return wayround_org.utils.range.get_range_first_index(offset, ranges)




import sys
import traceback
import pprint


def return_instant_exception_info(tb=False):
    return return_exception_info(
        sys.exc_info(),
        tb=tb
        )


def return_exception_info(exc_info, tb=False):
    txt = """
EXCEPTION: {type}
    VALUE: {val}
""".format(
        type=repr(exc_info[0].__name__),
        val=repr(exc_info[1])
        )

    if tb:
        txt += """
TRACEBACK:
{tb}
{feo}
    """.format(
            tb=''.join(
                traceback.format_list(traceback.extract_tb(exc_info[2]))
                ),
            feo=''.join(
                traceback.format_exception_only(exc_info[0], exc_info[1])
                )
            )

    return txt


def print_exception_info(e):
    txt = return_exception_info(e)
    print(txt)
    return


def prompt(g=None, l=None, prompt='-> ', readline_=True):

    if readline_:
        import readline

    bk = False

    while not bk:

        try:
            cmd = input(prompt)
            try:
                e = eval(cmd, g, l)
                if isinstance(e, dict):
                    pprint.pprint(e, indent=2)
                else:
                    print(repr(e))
            except SyntaxError:
                exec(cmd, g, l)

        except EOFError:
            bk = True
        except:
            print(
                return_exception_info(
                    sys.exc_info(),
                    tb=True
                    )
                )

    print('')

    return

extract_stack = traceback.extract_stack
format_list = traceback.format_list



"""
Check object type by searching it's attributes

This module supports two kinds of operation:
    1. Stranded - using original method of type recognition
    2. Native - using Python isinstance built-in function

Examples of execution:

import lxml.etree

import wayround_org.utils.types

>wayround_org.utils.types.types_n(lxml.etree.Element('q'))
['Hashable', 'Sized', 'Iterable', 'Container']

>wayround_org.utils.types.types_s(lxml.etree.Element('q'))
['Hashable', 'Sized', 'Mapping', 'Iterable', 'Container']
"""

import collections.abc

# This table is taken from python-3.3.2-docs-html/library/collections.abc.html

COMPARISON_TABLE = {
    'Container': {
        # Inherits from
        'i': [],
        # Abstract Methods
        'a': ['__contains__'],
        # Mixin Methods
        'm': []
        },
    'Hashable': {
        'i': [],
        'a': ['__hash__'],
        'm': []
        },
    'Iterable': {
        'i': [],
        'a': ['__iter__'],
        'm': []
        },
    'Iterator': {
        'i': ['Iterable'],
        'a': ['__next__'],
        'm': ['__iter__']
        },
    'Sized': {
        'i': [],
        'a': ['__len__'],
        'm': []
        },
    'Callable': {
        'i': [],
        'a': ['__call__'],
        'm': []
        },
    'Sequence': {
        'i': ['Sized', 'Iterable', 'Container'],
        'a': ['__getitem__', '__len__'],
        'm': ['__contains__', '__iter__', '__reversed__', 'index', 'count']
        },
    'MutableSequence': {
        'i': ['Sequence'],
        'a': ['__getitem__', '__setitem__', '__delitem__', '__len__',
              'insert'],
        'm': ['append', 'reverse', 'extend', 'pop', 'remove', '__iadd__']
        },
    'Set': {
        'i': ['Sized', 'Iterable', 'Container'],
        'a': ['__contains__', '__iter__', '__len__'],
        'm': ['__le__', '__lt__', '__eq__', '__ne__', '__gt__', '__ge__',
              '__and__', '__or__', '__sub__', '__xor__', 'isdisjoint']
        },
    'MutableSet': {
        'i': ['Set'],
        'a': ['__contains__', '__iter__', '__len__', 'add', 'discard'],
        'm': ['clear', 'pop', 'remove', '__ior__', '__iand__',
              '__ixor__', '__isub__']
        },
    'Mapping': {
        'i': ['Sized', 'Iterable', 'Container'],
        'a': ['__getitem__', '__iter__', '__len__'],
        'm': ['__contains__', 'keys', 'items', 'values', 'get', '__eq__',
              '__ne__']
        },
    'MutableMapping': {
        'i': ['Mapping'],
        'a': ['__getitem__', '__setitem__', '__delitem__', '__iter__',
              '__len__'],
        'm': ['pop', 'popitem', 'clear', 'update', 'setdefault']
        },
    }


def check_type_s(obj, name):
    """
    Check type using stranded method
    """

    if not name in COMPARISON_TABLE.keys():
        raise ValueError("Invalid type name")

    ret = True

    ref_type = COMPARISON_TABLE[name]

    for i in ref_type['i']:

        if not check_type_s(obj, i):
            ret = False
            break

    if ret:
        for i in ref_type['a']:
            if not hasattr(obj, i):
                ret = False
                break

    if ret:
        for i in ref_type['m']:
            if not hasattr(obj, i):
                ret = False
                break

    return ret


def types_s(obj):
    """
    Make list of types object in, using stranded method
    """

    ret = []

    for name in COMPARISON_TABLE.keys():
        if check_type_s(obj, name):
            ret.append(name)

    return ret


def check_type(obj, name):
    """
    Check type using native method
    """

    if not name in COMPARISON_TABLE.keys():
        raise ValueError("Invalid type name")

    return isinstance(obj, getattr(collections.abc, name))


def types(obj):
    """
    Make list of types object in, using native method
    """

    ret = []

    for name in COMPARISON_TABLE.keys():
        if check_type(obj, name):
            ret.append(name)

    return ret


for i in COMPARISON_TABLE.keys():
    exec(
        """
def is{name}_s(obj):
    return check_type_s(obj, '{name}')
""".format(name=i)
        )

del i

for i in COMPARISON_TABLE.keys():
    exec(
        """
def is{name}(obj):
    return check_type(obj, '{name}')
""".format(name=i)
        )

del i


STRUCT_CHECK_KEYS = ['t', 'te', 'None', '<', '>', '.', '', ' ', '{}']


def struct_check(value, struct):
    """
    Check whatever value corresponds to struct

    This implementation can't check dict with non-str keys, so exception is
    raised in this case

    Examples of `struct':


    {
    't': list, # type or tuple of types (like in isinstance). can also be a
               # string or tuple of strings with type names. see desc below
    'te': True, # exact type check with `type' function. default is True.
    'None': False, # can value be None? (False - default)
    '<': 0, # min child count. None - is default - no check
    '>': None, # max child count. None - is default - no check
    '.': { # check children. for work with sequences. default is None - not
           #                                                 check children
         't': list,
         '<': 1,
         '>': None,
         '.': {
              't': _Element
              }
         } # - default is None

    # works only if 't' is str or bytes
    '': False,
    ' ': False,  # '' - string emptiness allow or not,
                 # ' ' - only spaces in string allow or not

    '{}': {}, # if 't' is dict and only dict: True - any keys are possible and
              # each child checked with '.' value (recursion). False - same
              # as True, but without values checking (default)
              # if '{}' is dict, then all values in this dict must be dicts
              # with structures like this.
    }

    if 't' value is of type str, then value type is checked using this module's
    types() or types_s(): if '!' in beginning of 't' value, then types_s() is
    used

    return False if value does not matches struct, and True otherwise
    """

    ret = True

    if not isinstance(struct, dict):
        raise TypeError("`struct' must be dict")

    for i in list(struct.keys()):
        if not i in STRUCT_CHECK_KEYS:
            raise ValueError("invalid `struct' key: {}".format(i))

    typ = struct['t']

    if not isinstance(typ, tuple):
        typ = typ,

    iterable_type = types(value)

    # 'Iterable' required for checking sets
    iterable_type = 'Sequence' in iterable_type or 'Iterable' in iterable_type

    for i in typ:
        t_type = type(i)
        if t_type == str:
            if i.startswith('!'):
                if not i[1:] in COMPARISON_TABLE.keys():
                    raise ValueError("Invalid type name")
            else:
                if not i in COMPARISON_TABLE.keys():
                    raise ValueError("Invalid type name")

    type_exact = True
    if 'te' in struct:
        type_exact = struct['te']
        if not isinstance(type_exact, bool):
            raise TypeError("`te' must be bool")

    can_be_none = False
    if 'None' in struct:
        can_be_none = struct['None']
        if not isinstance(can_be_none, bool):
            raise TypeError("`None' must be bool")

    min_child_count = None
    if '<' in struct:
        min_child_count = struct['<']
        if min_child_count is not None and not isinstance(
                min_child_count, int):
            raise TypeError("`<' must be None or int")

    max_child_count = None
    if '>' in struct:
        max_child_count = struct['>']
        if max_child_count is not None and not isinstance(
                max_child_count, int):
            raise TypeError("`>' must be None or int")

    string_emptiness = True
    if '' in struct:
        string_emptiness = struct['']
        if not isinstance(string_emptiness, bool):
            raise TypeError("`' must be bool")

    string_is_space = True
    if ' ' in struct:
        string_is_space = struct[' ']
        if not isinstance(string_is_space, bool):
            raise TypeError("` ' must be bool")

    next_test = None
    if '.' in struct:
        if value is not None:
            next_test = struct['.']
            if next_test is not None and not isinstance(next_test, dict):
                raise TypeError("`.' must be None or dict")

            if next_test is not None and not iterable_type:
                raise ValueError(
                    "`.' is not None so value must be a sequence"
                    )

    dict_info = False
    if '{}' in struct:
        dict_info = struct['{}']
        if not type(dict_info) in [bool, dict]:
            raise TypeError(
                "`{{}}' must be bool or dict, but it's a: {}".format(
                    type(dict_info)
                    )
                )

        keys = list(dict_info.keys())

        for i in keys:
            if not isinstance(i, str):
                raise ValueError("keys in struct dict must be str")

            if not isinstance(dict_info[i], dict):
                raise ValueError("values in struct dict must be dict")

    if can_be_none == True and value == None:
        ret = True
    elif can_be_none == False and value == None:
        ret = False
    else:
        if ret:
            if type_exact == True:
                found = False
                for i in typ:
                    t_type = type(i)
                    if t_type == str:
                        if i.startswith('!'):
                            if i[1:] in types_s(value):
                                found = True
                                break
                        else:
                            if i in types(value):
                                found = True
                                break
                    else:
                        if isinstance(value, i):
                            found = True
                            break

                if not found:
                    ret = False
            else:
                if not isinstance(value, typ):
                    ret = False

        if ret:
            if min_child_count is not None:
                if len(value) < min_child_count:
                    ret = False

        if ret:
            if max_child_count is not None:
                if len(value) > max_child_count:
                    ret = False

        if ret:
            if (isinstance(value, str)
                    and value == ''
                    and string_emptiness == False):
                ret = False

        if ret:
            if (isinstance(value, bytes)
                    and value == b''
                    and string_emptiness == False):
                ret = False

        if ret:
            if (isinstance(value, str)
                    and value.isspace()
                    and string_is_space == False):
                ret = False

        if ret:
            if (isinstance(value, bytes)
                    and value.isspace()
                    and string_is_space == False):
                ret = False

        if ret:
            if len(typ) == 1 and typ[0] == dict:
                if isinstance(dict_info, dict):

                    keys = list(dict_info.keys())
                    for i in keys:
                        if not i in value:
                            ret = False
                            break
                        else:
                            ret = struct_check(value[i], dict_info[i])
                            if ret == False:
                                break

                    keys = list(value.keys())
                    for i in keys:
                        if not i in dict_info:
                            ret = False
                            break

                elif dict_info == False:
                    pass

                elif dict_info == True:
                    vkeys = list(value.keys())
                    for i in vkeys:
                        ret = struct_check(value[i], next_test)
                        if ret == False:
                            break
                else:
                    raise Exception("Programming error")

        if ret:
            if next_test is not None:
                if iterable_type:
                    for i in value:
                        if struct_check(i, next_test) == False:
                            ret = False
                            break

                else:
                    ret = False

    return ret


def attrs_object_to_dict(object_, dict_, attrs_keys=None):
    """
    attrs_keys: list of tuples, each of wich ('attr_name', 'key_name')
    """

    if attrs_keys is None:
        attrs_keys = list(object_.__dict__.keys())

    for i in attrs_keys:
        dict_[i[1]] = getattr(object_, i[0])

    return


def attrs_dict_to_object(
        dict_, object_, attrs_keys=None, create_new_attributes=False
        ):
    """
    attrs_keys: list of tuples, each of wich ('attr_name', 'key_name')
    """

    if attrs_keys is None:
        attrs_keys = list(dict_.keys())

    for i in attrs_keys:
        if hasattr(object_, i[0]) or create_new_attributes == True:
            setattr(object_, i[0], dict_[i[1]])
        else:
            raise KeyError(
                "object `{}' has no attribute `{}'".format(
                    object_,
                    i[0]
                    )
                )

    return


def attrs_dict_to_object_same_names(lst):

    ret = []

    for i in lst:
        ret.append((i, i,))

    return ret


def is_method(obj):
    return hasattr(obj, '__self__')


def value_to_bool(value):

    ret = False

    if isinstance(value, str):
        ret = text_to_bool(value)
    else:
        ret = bool(value)

    return ret


def text_to_bool(text):

    if not isinstance(text, str):
        raise TypeError("`text' must be str")

    return text.lower().strip() in ['1', 'yes', 'true', 'ok', 'y']


def is_all_rec_bytes_str(args, empty_fails=False, must_be=None):
    """
    Check's what all values in args ar bytes or str, recurcevely

    return bool, type (bool is True if passed. type is bytes, str or None)
    """

    if not isSequence(args):
        raise TypeError("`args' must be sequence, not {}".format(type(args)))

    if len(args) == 0:
        ret = not empty_fails, must_be
    else:

        tt = must_be
        error = False
        for i in args:

            if isinstance(i, list):
                res, res_t = is_all_rec_bytes_str(i, empty_fails, tt)

                if res == False:
                    error = True
                    break
                else:
                    if tt is None:
                        tt = res_t
                    else:
                        if tt != res_t:
                            error = True
                            break

            else:

                if tt is None:
                    tt = type(i)
                else:
                    if not isinstance(i, tt):
                        error = True
                        break

        if error:
            ret = False, None
        else:
            ret = True, tt

    return ret





from gi.repository import Gst, GLib


class SimpleSoundPlayer:

    def __init__(self):
        self._pipeline = None
        return

    def play(self, filename=None, data=None):
        if self.is_playing():
            self.stop()

        if filename:

            pipeline = Gst.Pipeline()
            self._pipeline = pipeline

            self._pipeline.set_name("audio-player")
            source = Gst.ElementFactory.make("filesrc", "file-source")
            demuxer = Gst.ElementFactory.make("oggdemux", "ogg-demuxer")
            decoder = Gst.ElementFactory.make("vorbisdec", "vorbis-decoder")
            conv = Gst.ElementFactory.make("audioconvert", "converter")
            sink = Gst.ElementFactory.make("alsasink", "audio-output")

            source.set_property('location', filename)

            demuxer.connect('pad-added', self._on_pad_added, decoder)

            pipeline.add(source)
            pipeline.add(demuxer)
            pipeline.add(decoder)
            pipeline.add(conv)
            pipeline.add(sink)

            source.link(demuxer)
            decoder.link(conv)
            conv.link(sink)

            self._pipeline.set_state(Gst.State.PLAYING)

        elif data:
            pipeline = Gst.Pipeline()
            self._pipeline = pipeline

            self._pipeline.set_name("audio-player")
            source = Gst.ElementFactory.make("appsrc", "data-source")
            demuxer = Gst.ElementFactory.make("oggdemux", "ogg-demuxer")
            decoder = Gst.ElementFactory.make("vorbisdec", "vorbis-decoder")
            conv = Gst.ElementFactory.make("audioconvert", "converter")
            sink = Gst.ElementFactory.make("alsasink", "audio-output")

            source.set_property('location', filename)

            demuxer.connect('pad-added', self._on_pad_added, decoder)

            pipeline.add(source)
            pipeline.add(demuxer)
            pipeline.add(decoder)
            pipeline.add(conv)
            pipeline.add(sink)

            source.link(demuxer)
            decoder.link(conv)
            conv.link(sink)

            self._pipeline.set_state(Gst.State.PLAYING)

        else:
            raise Exception("filename or data must be supplied")

        return

    def stop(self):
        self._pipeline.set_state(Gst.State.NULL)
        return

    def is_playing(self):
        return (
            self._pipeline
            and
            self._pipeline.get_state(0)[1] == Gst.State.PLAYING
            )

    def _on_pad_added(self, element, pad, decoder):
        sinkpad = decoder.get_static_pad('sink')
        pad.link(sinkpad)
        return


class SimpleSoundFileIdentifier:

    def __init__(self):
        return

    def identify(self, filename):
        pipeline = Gst.Pipeline()
        self._pipeline = pipeline
        pipeline.set_name("audio-player")
        source = Gst.ElementFactory.make("filesrc", "file-source")
        demuxer = Gst.ElementFactory.make("oggdemux", "ogg-demuxer")
        decoder = Gst.ElementFactory.make("vorbisdec", "vorbis-decoder")
        conv = Gst.ElementFactory.make("audioconvert", "converter")
        sink = Gst.ElementFactory.make("alsasink", "audio-output")

        source.set_property(
            'location',
            filename
            )

        demuxer.connect('pad-added', self._on_pad_added, decoder)

        pipeline.add(source)
        pipeline.add(demuxer)
        pipeline.add(decoder)
        pipeline.add(conv)
        pipeline.add(sink)

        source.link(demuxer)
        decoder.link(conv)
        conv.link(sink)
        return


def play(filename):
    ret = SimpleSoundFilePlayer()
    ret.play(filename)
    return ret


def identify(filename):
    ret = SimpleSoundFileIdentifier()
    ret = ret.identify(filename)
    return ret




import subprocess
import logging


def edit_file(filename, editor='emacs'):
    return edit_file_direct(filename, editor=editor)


def edit_file_direct(filename, editor='emacs'):
    p = None
    try:
        p = subprocess.Popen([editor, filename])
    except:
        logging.exception("error starting editor")
    else:
        try:
            p.wait()
        except:
            logging.exception("error waiting for editor")
        finally:
            if p.returncode is None:
                p.terminate()

        logging.info("editor exited")

    del(p)

    return




import ftplib
import logging
import os.path
import re

import wayround_org.utils.path


def v(s):
    logging.debug(s)


class FTPWalk:

    def __init__(self, ftp_connection, verbose=False):

        if not isinstance(ftp_connection, ftplib.FTP):
            raise TypeError

        self._ftpc = ftp_connection

        self._init(verbose)

    def __del__(self):
        self.clean()

    def _init(self, verbose=False):
        self._verbose = verbose

        self.re_ex = re.compile(
            r'(.*?)([ \t]+)(.*?)([ \t]+)(.*?)([ \t]+)(.*?)([ \t]+)(.*?)([ \t]+)(.*?)([ \t]+)(.*?)([ \t]+)(.*?)([ \t]+)(.*)')

        self.clean()

    def setverbose(self, value):
        self._verbose = value

    def clean(self):

        if self._verbose:
            v("Cleaning instance")

        self._file_lists = {}

        self._fstats_cb_dir = ''
        self._fstats_cb_errors = False

    def _fstats_cb(self, string):

        if not isinstance(string, str):
            raise TypeError

        if string == '':
            raise ValueError

        m = self.re_ex.match(string)

        if m is not None:

            self._file_lists[self._fstats_cb_dir][m.group(17)] = {
                'mode': m.group(1),
                'blocks': m.group(3),
                'uid': m.group(5),
                'gid': m.group(7),
                'size': m.group(9),
                'date1': m.group(11),
                'date2': m.group(13),
                'date3': m.group(15)
                }
        else:
            self._fstats_cb_errors = True

        del m

    def fstats(self, dirname, force=False):

        if self._verbose:
            v("Getting directory stats: {}".format(dirname))

        if self._ftpc is None:
            return 'not connected'

        if not isinstance(dirname, str):
            raise TypeError

        if dirname == '':
            raise ValueError

        if dirname in self._file_lists and not force:
            return 'ok'
        else:
            r = self.is_dir(dirname)

            if r == 'not connected':
                return 'not connected'

            elif r == 'errors':
                return 'errors'

            elif r == True:

                if dirname in self._file_lists:
                    del(self._file_lists[dirname])

                self._fstats_cb_dir = dirname
                self._file_lists[dirname] = {}
                self._fstats_cb_errors = False
                try:
                    self._ftpc.dir(dirname, self._fstats_cb)
                except:
                    self._fstats_cb_errors = True

                if self._fstats_cb_errors:
                    if dirname in self._file_lists:
                        del(self._file_lists[dirname])
                    return 'errors'
                else:
                    return 'ok'

            elif r == False:
                return 'not dir'

            else:
                return 'errors'

    def fstat_d_n(self, dirname, name):

        if self._ftpc is None:
            return 'not connected'

        if not isinstance(dirname, str):
            raise TypeError

        if dirname == '':
            raise ValueError

        if not isinstance(name, str):
            raise TypeError

        if name == '':
            raise ValueError

        if dirname in self._file_lists:
            if name in self._file_lists[dirname]:
                return self._file_lists[dirname][name]
            else:
                return 'file not found'
        else:
            r = self.fstats(dirname)

            if r == 'not connected':
                return 'not connected'

            elif r == 'ok':
                return self.fstat_d_n(dirname, name)

            elif r == 'not dir':
                return 'not dir'

            elif r == 'errors':
                return 'errors'

            else:
                return 'errors'

    def fstat(self, name):

        if self._ftpc is None:
            return 'not connected'

        if not isinstance(name, str):
            raise TypeError

        if name == '':
            raise ValueError

        name = wayround_org.utils.path.normpath(name)
        fdir = os.path.dirname(name)
        name = os.path.basename(name)
        return self.fstat_d_n(fdir, name)

    def _is(self, name, c='-', no=False):
        if self._ftpc is None:
            return 'not connected'

        if not isinstance(name, str):
            raise TypeError

        if name == '':
            raise ValueError

        if name == '/':
            return True

        r = self.fstat_d_n(
            os.path.dirname(name),
            os.path.basename(name)
            )

        if r == 'not connected':
            return 'not connected'

        elif r == 'file not found':
            return False

        elif r == 'error getting file stat':
            return 'errors'

        elif r == 'not dir':
            return False

        elif r == 'errors':
            return 'errors'

        if not no:
            if r['mode'][0] == c:
                return True
            else:
                return False
        else:
            if r['mode'][0] != c:
                return True
            else:
                return False

    def is_dir(self, name):
        return self._is(name, c='d')

    def is_not_dir(self, name):
        return self._is(name, c='d', no=True)

    def is_file(self, name):
        return self._is(name, c='-')

    def is_link(self, name):
        return self._is(name, c='l')

    def exists(self, name):

        if self._ftpc is None:
            return 'not connected'

        if not isinstance(name, str):
            raise TypeError

        if name == '':
            raise ValueError

        if name == '/':
            return True

        r = self.fstat_d_n(
            os.path.dirname(name),
            os.path.basename(name)
            )

        if r == 'not connected':
            return 'not connected'

        elif r == 'file not found':
            return False

        elif r == 'error getting file stat':
            return 'errors'

        elif r == 'not dir':
            return False

        elif r == 'errors':
            return 'errors'

        else:
            return True

    def get_all_files(self, dirname):

        if self._ftpc is None:
            return 'not connected'

        r = self.fstats(dirname)

        if r == 'ok':
            return self._file_lists[dirname]

        else:
            return r

    def get(self, dirname, c='d', no=False):
        l = []

        r = self.get_all_files(dirname)

        if not isinstance(r, dict):
            return r

        else:
            for i in r.keys():
                if no:
                    if r[i]['mode'][0] != c:
                        l.append(i)
                else:
                    if r[i]['mode'][0] == c:
                        l.append(i)
            return l

    def get_normal_files(self, dirname):
        return self.get(dirname, '-')

    def get_dirs(self, dirname):
        return self.get(dirname, 'd')

    def get_links(self, dirname):
        return self.get(dirname, 'l')

    def get_not_dirs(self, ftpc, dirname):
        return self.get(dirname, 'd', True)


class FTPWalkSole(FTPWalk):

    def __init__(self, server, verbose=False):

        if not isinstance(server, str):
            raise TypeError

        if server == '':
            raise ValueError

        self._server = server

        self._init()

    def connect(self):
        isok = True

        if self._verbose:
            v("Connecting to {}".format(self._server))

        if self._ftpc is not None:
            return True

        try:
            self._ftpc = ftplib.FTP(self._server)
        except:
            isok = False

        if isok:
            try:
                self._ftpc.login()
            except:
                isok = False

        if not isok:
            self.disconnect()

        if self._verbose:
            if not isok:
                v("Connection failed")

        return isok

    def disconnect(self):
        try:
            self._ftpc.quit()
        except:
            pass

        try:
            self._ftpc.close()
        except:
            pass

        self._ftpc = None

    def reconnect(self):
        self.disconnect()
        return self.connect()

    def clean(self):

        FTPWalk.clean(self)

        self.disconnect()

    # __del__ not required as it is present in parent Class




import io
import logging
import os.path
import subprocess
import sys
import threading
import time

import wayround_org.utils.stream


def simple_exec(
        program,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
        options=None,
        bufsize=(2 * 1024 ** 2),
        cwd=None
        ):

    if options is None:
        options = []

    p = None

    try:
        p = subprocess.Popen(
            [program] + options,
            stdin=stdin, stdout=stdout, stderr=stderr,
            bufsize=bufsize,
            cwd=cwd
            )
    except:
        logging.exception("Error starting `{}' subprocess".format(program))
        p = None
        raise

    return p


def pipe_subprocesses(
        processes_list,
        processes_names,
        bufsize=(2 * 1024 ** 2),
        verbose=False
        ):

    if not isinstance(processes_list, list):
        raise ValueError("processes_list must be a list")

    for i in processes_list:
        if not isinstance(i, subprocess.Popen):
            raise ValueError("processes_list must be a list of Popen objects")

    if not isinstance(processes_names, list):
        raise ValueError("processes_names must be a list")

    for i in processes_names:
        if not isinstance(i, str):
            raise ValueError("processes_list must be a list of strings")

    if len(processes_list) != len(processes_names):
        raise ValueError("len(processes_list) != len(processes_names)")

    count = len(processes_list)
    cats = []
    for i in range(count - 1):

        if verbose:
            thread_name = "{} >> {}".format(
                processes_names[i],
                processes_names[i + 1]
                )
        else:
            thread_name = 'Thread'

        logging.debug("Creating thread `{}'".format(thread_name))

        cats.append(
            wayround_org.utils.stream.cat(
                processes_list[i].stdout,
                processes_list[i + 1].stdin,
                threaded=True,
                close_output_on_eof=True,
                bs=bufsize,
                thread_name=thread_name,
                apply_input_seek=False,
                apply_output_seek=False
                )
            )

    for i in cats:
        i.start()

    for i in range(count - 1):
        cats[i].join()
        processes_list[i].wait()

    processes_list[-1].wait()

    return


def test_pipes():

    for i in [
            (logging.CRITICAL, '-c-'),
            (logging.ERROR, '-e-'),
            (logging.WARN, '-w-'),
            (logging.WARNING, '-w-'),
            (logging.INFO, '-i-'),
            (logging.DEBUG, '-d-')
            ]:
        logging.addLevelName(i[0], i[1])

    logging.basicConfig(level='DEBUG')

    f = open('/proc/cpuinfo', 'rb')
    text = f.read()
    f.close()

    # =========== xz based testing ===========

    txt = io.BytesIO()
    txt.write(text)
    txt.seek(0)

    xz1 = simple_exec(
        'xz',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        options=['-9'],
        bufsize=0
        )

    xz2 = simple_exec(
        'xz',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        options=['-d'],
        bufsize=0
        )

    xz3 = simple_exec(
        'xz',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        options=['-9'],
        bufsize=0
        )

    xz4 = simple_exec(
        'xz',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        options=['-d'],
        bufsize=0
        )

    cat1 = wayround_org.utils.stream.cat(
        txt,
        xz1.stdin,
        threaded=True,
        close_output_on_eof=True,
        bs=200,
        thread_name="File >> xz1"
        )

    cat2 = wayround_org.utils.stream.cat(
        xz4.stdout,
        sys.stdout,
        threaded=True,
        close_output_on_eof=False,
        convert_to_str='utf-8',
        bs=200,
        thread_name="xz4 >> STDOUT"
        )

    cat1.start()
    cat2.start()

    pipe_subprocesses(
        [xz1, xz2, xz3, xz4],
        ['xz1', 'xz2', 'xz3', 'xz4'],
        bufsize=2 * 1024 ** 2,
        verbose=True
        )

    cat1.join()
    cat2.join()

    # =========== bzip2 based testing ===========

#    txt = None
#    txt = io.BytesIO()
#    txt.write(text)
    txt.seek(0)

    bzip21 = simple_exec(
        'bzip2',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        options=['-9'],
        bufsize=0
        )

    bzip22 = simple_exec(
        'bzip2',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        options=['-d'],
        bufsize=0
        )

    bzip23 = simple_exec(
        'bzip2',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        options=['-9'],
        bufsize=0
        )

    bzip24 = simple_exec(
        'bzip2',
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        options=['-d'],
        bufsize=0
        )

    cat1 = wayround_org.utils.stream.cat(
        txt,
        bzip21.stdin,
        threaded=True,
        close_output_on_eof=True,
        bs=200,
        thread_name="File >> bzip2_1"
        )

    cat2 = wayround_org.utils.stream.cat(
        bzip24.stdout,
        sys.stdout,
        threaded=True,
        close_output_on_eof=False,
        convert_to_str='utf-8',
        bs=200,
        thread_name="bzip2_4 >> STDOUT"
        )

    cat1.start()
    cat2.start()

    pipe_subprocesses(
        [bzip21, bzip22, bzip23, bzip24],
        ['bzip2_1', 'bzip2_2', 'bzip2_3', 'bzip2_4'],
        bufsize=2 * 1024 ** 2,
        verbose=True
        )

    cat1.join()
    cat2.join()

    txt.close()


class ProcessStream:

    def __init__(self):

        self.clear(init=True)

    def clear(self, init=False):

        ret = 0

        if not init and self.stat() != 'stopped':
            logging.error(
                "Program settings can not be cleaned while it is working"
                )

            ret = 1
        else:

            if not init:
                self.stop()

            self.program = None
            self.stdin = None
            self.stdout = None
            self.stderr = None
            self.options = None
            self.proc_bufsize = None
            self.cat_bufsize = None
            self.cwd = None
            self.verbose = None
            self.close_output_on_eof = None
            self.flush_on_input_eof = None
            self.stdin_mode = None
            self.stdout_mode = None

            self.proc = None
            self._wait_thread = None

            self.in_cat = None
            self.out_cat = None

            if not init:
                if self.stop_flag:
                    self.stop_flag.set()

            self.stop_flag = threading.Event()

            if init:
                self.returncode = None

        return ret

    def setValues(
            self,
            program,
            stdin,
            stdout,
            stderr,
            options=None,
            proc_bufsize=0,
            cat_bufsize=(2 * 1024 ** 2),
            cwd=None,
            verbose=False,
            close_output_on_eof=False,
            flush_on_input_eof=False,
            stdin_mode='sync',
            stdout_mode='sync'
            ):

        if options is None:
            options = []

        if not stdin_mode in ['sync', 'async']:
            raise ValueError("stdin_mode must be in ['sync', 'async']")

        if not stdout_mode in ['sync', 'async']:
            raise ValueError("stdout_mode must be in ['sync', 'async']")

        ret = 0

        if self.stat() != 'stopped':
            logging.error(
                "Program settings can not be changed while it is working"
                )

            ret = 1

        else:

            self.program = program
            self.stdin = stdin
            self.stdout = stdout
            self.stderr = stderr
            self.options = options
            self.proc_bufsize = proc_bufsize
            self.cat_bufsize = cat_bufsize
            self.cwd = cwd
            self.verbose = verbose
            self.close_output_on_eof = close_output_on_eof
            self.flush_on_input_eof = flush_on_input_eof
            self.stdin_mode = stdin_mode
            self.stdout_mode = stdout_mode

        return ret

    def start(self):

        ret = 0

        if self.stat() != 'stopped':
            logging.error(
                "Program can not be started if it's already working"
                )

            ret = 1

        if ret == 0:

            try:
                self.proc = simple_exec(
                    self.program,
                    stdin=subprocess.PIPE,
                    stdout=subprocess.PIPE,
                    stderr=self.stderr,
                    options=self.options,
                    bufsize=self.proc_bufsize,
                    cwd=self.cwd
                    )
            except:
                logging.exception(
                    "Error starting process `{}'".format(self.program)
                    )
                ret = 2

        if ret == 0:
            try:
                self._wait_thread = threading.Thread(
                    name="Thread Waiting for program `{}' exit".format(
                        self.program
                        ),
                    target=self._proc_waiter
                    )
                self._wait_thread.start()
            except:
                logging.exception("Error starting thread")
                ret = 3

        if ret == 0:

            try:
                thread_name = ''

                if self.verbose:
                    thread_name = 'in >> {}'.format(self.proc.pid)
                else:
                    thread_name = 'Thread'

                # TODO: add error handlers

                self.in_cat = wayround_org.utils.stream.cat(
                    self.stdin,
                    self.proc.stdin,
                    threaded=True,
                    flush_on_input_eof=True,
                    close_output_on_eof=True,
                    bs=self.cat_bufsize,
                    thread_name=thread_name,
                    verbose=self.verbose,
                    termination_event=self.stop_flag,
                    on_exit_callback=self._close_in_cat,
                    on_input_read_error=None,
                    on_output_write_error=None,
                    read_type=self.stdin_mode,
                    write_type='sync',
                    apply_input_seek=False,
                    apply_output_seek=False
                    )

                if self.verbose:
                    thread_name = '{} >> out'.format(self.proc.pid)
                else:
                    thread_name = 'Thread'

                self.out_cat = wayround_org.utils.stream.cat(
                    self.proc.stdout,
                    self.stdout,
                    threaded=True,
                    flush_on_input_eof=self.flush_on_input_eof,
                    close_output_on_eof=self.close_output_on_eof,
                    bs=self.cat_bufsize,
                    thread_name=thread_name,
                    verbose=self.verbose,
                    termination_event=self.stop_flag,
                    on_exit_callback=self._close_out_cat,
                    on_input_read_error=None,
                    on_output_write_error=None,
                    read_type='sync',
                    write_type=self.stdout_mode,
                    apply_input_seek=False,
                    apply_output_seek=False
                    )

                self.in_cat.start()
                self.out_cat.start()

            except:
                logging.exception(
                    "Some exception while starting cat threads"
                    )
                ret = 4

        if ret != 0:
            self.stop()

        return ret

    def stop(self):

        logging.debug("Signaled stop for `{}'".format(self.program))

        if self.stat() != 'stopped':

            self.stop_flag.set()

            if self.in_cat:
                self.in_cat.join()

            if self.out_cat:
                self.out_cat.join()

            self.wait('stopped')

            self.clear()

        return

    def stat(self):

        ret = 'unknown'

        v1 = self.in_cat
        v2 = self.out_cat
        v3 = self._wait_thread

        if (v1 is not None
                and v2 is not None
                and v3 is not None):
            ret = 'working'

        if (v1 is None
                and v2 is None
                and v3 is None):
            ret = 'stopped'

        logging.debug("""\
status (for `{}'):
self.in_cat     {}
self.out_cat    {}
self.proc       {}
{}
""".format(self.program, self.in_cat, self.out_cat, self.proc, ret))

        return ret

    def wait(self, what='stopped'):

        # TODO: add timeout

        if not what in ['stopped', 'working']:
            raise ValueError("wrong `what' value")

        while True:

            logging.debug("Waiting for status `{}'".format(what))

            s = self.stat()

            if s == what:
                break

            time.sleep(0.2)

        return self.returncode

    def _close_in_cat(self):
        self.in_cat.join()
        self.in_cat = None
        return

    def _close_out_cat(self):
        self.out_cat.join()
        self.out_cat = None
        return

    def _proc_waiter(self):

        while True:

            try:
                self.proc.wait(0.2)
            except subprocess.TimeoutExpired:
                pass
            else:
                self.returncode = self.proc.returncode
                break

        self._wait_thread = None

        return


def process_stream(
        program,
        stdin,
        stdout,
        stderr,
        options=None,
        proc_bufsize=(2 * 1024 ** 2),
        cat_bufsize=(2 * 1024 ** 2),
        cwd=None,
        verbose=False,
        close_output_on_eof=True,
        flush_on_input_eof=True,
        stdin_mode='sync',
        stdout_mode='sync'
        ):

    if options is None:
        options = []

    ps = ProcessStream()

    ps.setValues(
        program=program,
        stdin=stdin,
        stdout=stdout,
        stderr=stderr,
        options=options,
        proc_bufsize=proc_bufsize,
        cat_bufsize=cat_bufsize,
        cwd=cwd,
        verbose=verbose,
        close_output_on_eof=close_output_on_eof,
        flush_on_input_eof=flush_on_input_eof,
        stdin_mode=stdin_mode,
        stdout_mode=stdout_mode
        )

    res = ps.start()

    logging.debug("ps.start() == {}".format(res))

    ret = ps.wait()

    logging.debug("ps.wait() == {}".format(ret))

    if ret is None:
        ret = 222

    return ret


def process_file(
        program,
        infile,
        outfile,
        stderr=subprocess.DEVNULL,
        options=None,
        cwd=None,
        bufsize=(2 * 1024 ** 2),
        verbose=False
        ):

    if options is None:
        options = []

    ret = 0

    if not os.path.isfile(infile):
        logging.error("Input file not exists: {}".format(infile))
        ret = 1
    else:
        try:
            fi = open(infile, 'rb')
        except:
            logging.exception("Can't open file `{}'".format(infile))
        else:

            try:
                fo = open(outfile, 'wb')
            except:
                logging.exception("Can't rewrite file `{}'".format(outfile))
            else:

                try:
                    options = []

                    ec = process_stream(
                        program,
                        fi,
                        fo,
                        stderr=stderr,
                        options=options,
                        cwd=cwd,
                        verbose=verbose,
                        close_output_on_eof=True,
                        flush_on_input_eof=True,
                        stdin_mode='sync',
                        stdout_mode='sync'
                        )

                    if ec != 0:
                        logging.error(
                            "Error processing file {in} to {out} through"
                            " {proc} (code {code})".format_map(
                                {
                                    'in': infile,
                                    'out': outfile,
                                    'proc': program,
                                    'code': ec
                                    }
                                )
                            )
                        ret = 2
                finally:
                    fo.close()

            finally:
                fi.close()

    return ret



#!/usr/bin/python3

import subprocess
import os.path

from setuptools import setup
from setuptools import Extension


py_compil_args = None
py_link_args = None

p = subprocess.Popen(
    ['pkg-config', '--cflags', 'python3'], stdout=subprocess.PIPE)
p.wait()
py_compile_args = str(p.communicate()[0], encoding='utf-8').split()

p = subprocess.Popen(
    ['pkg-config', '--libs', 'python3'], stdout=subprocess.PIPE)
p.wait()
py_link_args = str(p.communicate()[0], encoding='utf-8').split()




setup(
    name='wayround_org_utils',
    version='1.10.3',
    description='Various service modules',
    long_description="""\
This package contains various useful modules functions and classes.

""",

    author='Alexey Gorshkov',
    author_email='animus@wayround.org',
    url='https://github.com/AnimusPEXUS/wayround_org_utils',
    install_requires=[
        'wayround_org_mail',
        'regex'
        ],
    packages=[
        'wayround_org.utils',
        'wayround_org.utils.format'
        ],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX',
        ],
    ext_modules=[
        Extension(
            "wayround_org.utils.format.elf_bin",
            ["wayround_org/utils/format/elf_bin.c"],
            extra_compile_args=py_compile_args,
            extra_link_args=py_link_args,
            ),
        ] ,
    package_data={
        'wayround_org.utils': [
            'config.sub',
            os.path.join('format', '*.c'),
            os.path.join('format', '*.h')
            ]
        }
    )




"""
Based on RFC5322
"""


import re
import datetime

import wayround_org.utils.datetime_iso8601


DATE_EXPRESSION = (
    r'\s*'
    r'(?P<day>\d+)'
    r'\s+'
    r'(?P<month>[a-zA-Z]{3})'
    r'\s+'
    r'(?P<year>\d+)'
    r'\s*'
    )

TIMEOFDAY_EXPRESSION = (
    r'\s*'
    r'(?P<hour>\d+)'
    r'\:(?P<minute>\d+)'
    r'(\:(?P<second>\d+))?'
    r'\s*'
    )

ZONE_EXPRESSION = (
    r'\s*'
    r'(?P<zone_sign>[+-]?)'
    r'(?P<zone_val>(?P<zone_hours>\d{2})(?P<zone_minutes>\d{2}))'
    r'\s*'
    )

TIME_EXPRESSION = (
    r'\s*'
    r'(?P<timeofday>{timeofday})'
    r'\s+'
    r'(?P<zone>{zone})'
    r'\s*'.format(
        timeofday=TIMEOFDAY_EXPRESSION,
        zone=ZONE_EXPRESSION
        )
    )

DATETIME_EXPRESSION = (
    r'\s*'
    r'((?P<dayofweek>[a-zA-Z]{{3}})\,)?'
    r'\s+'
    r'(?P<date>{date})'
    r'\s+'
    r'(?P<time>{time})'
    r'\s*'
    r'').format(
    date=DATE_EXPRESSION,
    time=TIME_EXPRESSION
    )


DATETIME_EXPRESSION_C = re.compile(DATETIME_EXPRESSION)

MONTHES = [
    "Jan", "Feb", "Mar", "Apr",
    "May", "Jun", "Jul", "Aug",
    "Sep", "Oct", "Nov", "Dec"
    ]

DAYSOFWEEK = [
    "Mon", "Tue", "Wed", "Thu",
    "Fri", "Sat", "Sun"
    ]


def str_to_datetime(value):
    """
    Parse string and make datetime.datetime of it
    """

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    ret = None

    ret_attributes = set()
    re_res = DATETIME_EXPRESSION_C.match(value)

    if re_res:

        month = MONTHES.index(re_res.group('month')) + 1

        second = re_res.group('second')
        if second is None:
            second = 0
        else:
            second = int(second)
            ret_attributes.add('second')

        zonesign = '+'
        if re_res.group('zone_sign') == '-':
            zonesign = '-'

        tzinfo = wayround_org.utils.datetime_iso8601.gen_tz(
            int(re_res.group('zone_hours')),
            int(re_res.group('zone_minutes')),
            plus=zonesign == '+'
            )

        result = datetime.datetime(
            year=int(re_res.group('year')),
            month=month,
            day=int(re_res.group('day')),
            hour=int(re_res.group('hour')),
            minute=int(re_res.group('minute')),
            second=second,
            tzinfo=tzinfo
            )

        if re_res.group('dayofweek') is not None:
            ret_attributes.add('dayofweek')

        ret = result

    return ret, ret_attributes


def datetime_to_str(value, attrs=None):

    if not isinstance(value, datetime.datetime):
        raise TypeError("`value' must be inst of datetime.datetime")

    if not hasattr(value, 'tzinfo') or value.tzinfo is None:
        raise ValueError("supplied datetime must have time zone info")

    if attrs is None:
        attrs = set(['dayofweek', 'second'])

    dayofweek = ''
    if 'dayofweek' in attrs:
        dayofweek = '{}, '.format(DAYSOFWEEK[value.date().weekday()])

    second = ''
    if 'second' in attrs:
        second = ':{:02d}'.format(value.second)

    month = MONTHES[value.month - 1]

    zone = wayround_org.utils.datetime_iso8601.format_tz(
        value.tzinfo,
        sep=False
        )

    ret = (
        '{dayofweek}'
        '{day:02d} {month} {year:04d} '
        '{hour:02d}:{minute:02d}{second} {zone}'
        ).format(
            dayofweek=dayofweek,
            day=value.day,
            month=month,
            year=value.year,
            hour=value.hour,
            minute=value.minute,
            second=second,
            zone=zone
            )

    return ret


def str_parse_test():
    """
    run tests if datetime string parser
    """

    tests = [
        'Fri, 29 Jan 2016 13:36:08 +0200'
        ]

    for value in tests:

        print("test subject: {}".format(value))

        re_res = DATETIME_EXPRESSION_C.match(value)

        if re_res:

            for i in [
                    'dayofweek',
                    'day', 'month', 'year',
                    'hour', 'minute', 'second',
                    'zone_sign', 'zone_val',
                    'zone_hours', 'zone_minutes']:
                print("    {:20}: {}".format(i, re_res.group(i)))

            dt, attrs = str_to_datetime(value)

            print(
                "    {:20}: {}".format(
                    'iso8601',
                    wayround_org.utils.datetime_iso8601.datetime_to_str(
                        dt
                        )
                    )
                )

            print(
                "    {:20}: {}".format(
                    'rendered',
                    datetime_to_str(
                        dt
                        )
                    )
                )
        else:
            print("    parse error")
    return





"""
Based on ISO 8601-2004
Partially supported truncated dates and times
"""

import datetime
import re

# FIXME: Year expansions are not supported due to python datetime limitations


DATE_EXPRESSION = (
    r'^'
    r'(?P<year>((\d{4})|(\d{2})))'
    r'(?P<not_year>(?P<sep1>\-)?(?P<month>\d{2})((?P<sep2>\-)?'
    r'(?P<day>\d{2}))?)?'
    r'$'
    )

DATE_EXPRESSION_C = re.compile(DATE_EXPRESSION)

DATE_TRUNCATED_EXPRESSION = (
    r'^'
    r'--'
    r'((?P<number1>\d{2})|-)(?P<sep>-)?(?P<number2>\d{2})?'
    r'$'
    )

DATE_TRUNCATED_EXPRESSION_C = re.compile(DATE_TRUNCATED_EXPRESSION)

DATE_ORDINAL_EXPRESSION = (
    r'^'
    r'(?P<year>\d{4})(?P<sep>\-)?(?P<day>\d{3})'
    r'$'
    )

DATE_ORDINAL_EXPRESSION_C = re.compile(DATE_ORDINAL_EXPRESSION)

DATE_WEEK_EXPRESSION = (
    r'^'
    r'(?P<year>\d{4})(?P<sep1>\-)?'
    r'W(?P<week>\d{2})((?P<sep2>\-)?(?P<day>\d{1}))?'
    r'$'
    )

DATE_WEEK_EXPRESSION_C = re.compile(DATE_WEEK_EXPRESSION)

TIME_EXPRESSION = (
    r'^'
    r'(?P<T>T)?(?P<hour>\d{2})((?P<sep1>\:)?(?P<min>\d{2})((?P<sep2>\:)?'
    r'(?P<sec>\d{2}))?)?'
    r'((?P<fract_sep>[\.\,])(?P<fract>\d+))?'
    r'(?P<tz>'
    r'((?P<Z>Z)|(?P<tz_sign>[+-])(?P<tz_hour>\d{2})(?P<sep3>\:)?'
    r'(?P<tz_min>\d{2})))?'
    r'$'
    )

TIME_EXPRESSION_C = re.compile(TIME_EXPRESSION)

TIME_TRUNCATED_EXPRESSION = (
    r'^'
    r'-'
    r'((?P<number1>\d{2})|-)(?P<sep>:)?(?P<number2>\d{2})?'
    r'$'
    )

TIME_TRUNCATED_EXPRESSION_C = re.compile(TIME_TRUNCATED_EXPRESSION)


DATE_ATTRIBUTES = {
    '-', 'year', 'century', 'week', 'ordinal', 'day', 'month', 'truncated_date'
    }
TIME_ATTRIBUTES = {
    'hour', 'min', 'sec', 'fract', 'tz_hour', 'tz_min', 'fract_sep',
    'T', ':', 'local', 'utc', ',', '.', 'Z', 'truncated_time'
    }

DEFAULT_DATE_ATTRIBUTES = {
    '-', 'year', 'day', 'month'
    }

DEFAULT_TIME_ATTRIBUTES = {
    'hour', 'min', 'sec', 'fract', 'tz_hour', 'tz_min', 'fract_sep',
    'T', ':', '.'
    }


def _date_attrs_check(arr):
    for i in arr:
        if not i in DATE_ATTRIBUTES:
            raise ValueError("invalid date attribute `{}'".format(i))
    return


def _time_attrs_check(arr):
    for i in arr:
        if not i in TIME_ATTRIBUTES:
            raise ValueError("invalid time attribute `{}'".format(i))
    return


def str_to_date(value):

    if not isinstance(value, str):
        raise TypeError("`value' must be str")

    res, attr = date_truncated_str_to_dict(value)
    if res is not None:
        ret = res, attr
    else:
        res, attr = date_ordinal_str_to_date(value)
        if res is not None:
            ret = res, attr
        else:
            res, attr = date_week_str_to_date(value)
            if res is not None:
                ret = res, attr
            else:
                res, attr = date_normal_str_to_date(value)
                if res is not None:
                    ret = res, attr
                else:
                    ret = None, None

    return ret


def str_to_time(value):

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    if value[0] == '-':
        # TODO: separate this
        ret, ret_attributes = time_truncated_str_to_dict(value)
    else:

        _debug = False

        ret = None
        ret_attributes = set()

        res = TIME_EXPRESSION_C.match(value)

        if res is None:
            pass
        else:
            groupdict = res.groupdict()
            separator_error = False
            separator = groupdict['sep1']
            microseconds = 0

            if groupdict['T'] == 'T':
                ret_attributes.add('T')

            if separator == ':':
                ret_attributes.add(':')

            if (groupdict['sec'] is not None
                    and groupdict['sep2'] != separator):
                separator_error = True

            if (groupdict['tz_min'] is not None
                    and groupdict['sep3'] != separator):
                separator_error = True

            if groupdict['fract_sep'] is not None:
                ret_attributes.add(groupdict['fract_sep'])

            for i in ['hour', 'min', 'sec', 'fract', 'tz_hour', 'tz_min']:
                if groupdict[i] is not None:
                    ret_attributes.add(i)

            if groupdict['fract'] is not None:
                fract = int(groupdict['fract'])

                if groupdict['min'] is None:
                    minute = float(60 * float('0.{}'.format(fract)))
                    groupdict['min'] = str(int(minute))
                    fract = int(str(minute).split('.')[1])

                if groupdict['sec'] is None:
                    second = float(60 * float('0.{}'.format(fract)))
                    groupdict['sec'] = str(int(second))
                    fract = int(str(second).split('.')[1])

                microseconds = round(1000000 * float('0.{}'.format(fract)))

                if _debug:
                    print("After fract translations:\n{}".format(groupdict))
                    print("microseconds = {}".format(microseconds))

            for i in ['hour', 'min', 'sec', 'fract', 'tz_hour', 'tz_min']:
                if groupdict[i] is None:
                    groupdict[i] = '00'

            if groupdict['tz_sign'] is None:
                groupdict['tz_sign'] = '+'

            if separator_error:
                pass
            else:
                z = None
                if groupdict['tz'] is None:
                    ret_attributes.add('local')
                    z = None
                elif groupdict['tz'] == 'Z':
                    ret_attributes.add('Z')
                    ret_attributes.add('utc')
                    z = datetime.timezone.utc
                else:
                    ret_attributes.add('utc')
                    z = gen_tz(
                        int(groupdict['tz_hour']),
                        int(groupdict['tz_min']),
                        groupdict['tz_sign'] == '+'
                        )

                if _debug:
                    print("Setting tz to {}".format(z))

                ret = datetime.time(
                    int(groupdict['hour']),
                    int(groupdict['min']),
                    int(groupdict['sec']),
                    microseconds,
                    tzinfo=z
                    )

                _time_attrs_check(ret_attributes)

    return ret, ret_attributes


def str_to_datetime(value):

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    _debug = False
    ret = None
    ret_attributes = set()
    t_index = value.find('T')
    if t_index != -1:
        date_str = value[:t_index]
        time_str = value[t_index:]

        if _debug:
            print(
                "date part: '{}', time part: '{}'".format(date_str, time_str)
                )

        date, d_attrs = str_to_date(date_str)
        time, t_attrs = str_to_time(time_str)

        if _debug:
            if date is None:
                print("Date not parsed")
            if time is None:
                print("Time not parsed")

        if date is not None and time is not None:

            date_can_have_separators = (
                'month' in d_attrs
                or 'day' in d_attrs
                or 'week' in d_attrs
                )

            time_can_have_separators = (
                'min' in t_attrs
                or 'tz_min' in t_attrs
                )

            separator_error = False

            if date_can_have_separators and time_can_have_separators:
                if '-' in d_attrs and ':' in t_attrs:
                    pass
                else:
                    separator_error = True

            if not separator_error:
                ret = datetime.datetime.combine(date, time)
                ret_attributes = d_attrs | t_attrs

    # else:
    #    res, attr = str_to_date(value)
    #
    #    if res != None:
    #        ret = res

    return ret, ret_attributes


def date_to_str(date, attr=None):

    if not isinstance(date, datetime.date):
        raise TypeError("`date' must be inst of datetime.date")

    if attr is None:
        attr = DEFAULT_DATE_ATTRIBUTES

    if 'truncated_date' in attr:

        ret = dict_to_truncated_date_str(date, attr)

    elif 'ordinal' in attr:

        cop = datetime.date(date.year, 1, 1)

        year = '{:04d}'.format(date.year)
        day = ''
        sep = ''

        if 'day' in attr and '-' in attr:
            sep = '-'

        if 'day' in attr:
            day = '{:03d}'.format((date - cop).days)

        ret = '{year}{sep}{day}'.format(
            year=year,
            sep=sep,
            day=day
            )

    elif 'week' in attr:

        cop = datetime.date(date.year, 1, 1)
        delta = date - cop
        weeks = int(delta.days / 7)
        days = delta.days - (weeks * 7)
#        days = 7 * float('0.{}'.format(int(str(weeks).split('.')[1])))

        year = '{:04d}'.format(date.year)
        week = '{:02d}'.format(weeks)
        day = ''
        sep1 = ''
        sep2 = ''

        if 'day' in attr:
            day = '{:01d}'.format(int(days))

        if '-' in attr:
            sep1 = '-'

        if 'day' in attr and '-' in attr:
            sep2 = '-'

        ret = '{year}{sep1}W{week}{sep2}{day}'.format(
            year=year,
            week=week,
            day=day,
            sep1=sep1,
            sep2=sep2
            )

    else:

        year = '{:04d}'.format(date.year)
        month = ''
        day = ''
        sep1 = ''
        sep2 = ''

        if 'century' in attr:
            year = year[:-2]

        if 'month' in attr and '-' in attr:
            sep1 = '-'

        if 'month' in attr:
            month = '{:02d}'.format(date.month)

        if 'day' in attr and '-' in attr:
            sep2 = '-'

        if 'day' in attr:
            day = '{:02d}'.format(date.day)

        ret = '{year}{sep1}{month}{sep2}{day}'.format(
            year=year,
            month=month,
            day=day,
            sep1=sep1,
            sep2=sep2
            )

    return ret


def time_to_str(time, attr=None):

    if not isinstance(time, datetime.time):
        raise TypeError("`time' must be inst of datetime.time")

    if 'truncated_time' in attr:
        ret = dict_to_truncated_time_str(time, attr)
    else:

        if attr is None:
            attr = DEFAULT_TIME_ATTRIBUTES

        t = ''
        if 'T' in attr:
            t = 'T'

        hour = '{:02d}'.format(time.hour)

        fract = 0
        if time.microsecond != 0:
            fract = time.microsecond / 1000000

        second = 0
        if 'sec' in attr:
            second = '{:02d}'.format(time.second)
        else:
            second = float(time.second + fract)
            fract = second / 60
            second = ''

        minute = 0
        if 'min' in attr:
            minute = '{:02d}'.format(time.minute)
        else:
            minute = float(time.minute + fract)
            fract = minute / 60
            minute = ''

        fract_sep = ''
        for i in [',', '.']:
            if i in attr:
                fract_sep = i
                break

        sep1 = ''
        if 'min' in attr and ':' in attr:
            sep1 = ':'

        sep2 = ''
        if 'sec' in attr and ':' in attr:
            sep2 = ':'

        if not 'fract' in attr:
            fract = ''
        else:
            splitted_fract = str(fract).split('.')
            if len(splitted_fract) > 1:
                fract = str(splitted_fract[1])
            else:
                fract = '0'

        tz = format_tz(
            time.tzinfo,
            sep=':' in attr,
            minu='tz_min' in attr,
            zed='Z' in attr
            )

        ret = \
            '{T}{hour}{sep1}{min}{sep2}' \
            '{sec}{fract_sep}{fract}{tz}'.format(
                T=t,
                sep1=sep1,
                sep2=sep2,
                fract_sep=fract_sep,
                hour=hour,
                min=minute,
                sec=second,
                fract=fract,
                tz=tz
                )

    return ret


def datetime_to_str(value, attr=None):

    if not isinstance(value, datetime.datetime):
        raise TypeError("`value' must be inst of datetime.datetime")

    if attr is None:
        attr = DEFAULT_DATE_ATTRIBUTES | DEFAULT_TIME_ATTRIBUTES

    date = value.date()
    time = value.timetz()

    date_str = date_to_str(date, attr)
    time_str = time_to_str(time, attr)

    t = ''
    if not time_str[0] == 'T':
        t = 'T'

    ret = '{date}{t}{time}'.format(
        date=date_str,
        time=time_str,
        t=t
        )

    return ret


def date_truncated_str_to_dict(value):

    if not isinstance(value, str):
        raise TypeError("`value' must be inst if str")

    _debug = False

    ret = None
    ret_attributes = set()

    res = DATE_TRUNCATED_EXPRESSION_C.match(value)

    if res is None:
        if _debug:
            print("no match")
    else:

        ret = {'month': None, 'day': None}

        ret_attributes.add('truncated_date')
        groupdict = res.groupdict()

        if groupdict['number1'] not in [None, '-']:
            ret_attributes.add('month')
            ret['month'] = int(groupdict['number1'])

        if groupdict['number2'] not in [None]:
            ret_attributes.add('day')
            ret['day'] = int(groupdict['number2'])

        if groupdict['sep'] == '-':
            ret_attributes.add('-')

    return ret, ret_attributes


def time_truncated_str_to_dict(value):

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    _debug = False

    ret = None
    ret_attributes = set()

    res = TIME_TRUNCATED_EXPRESSION_C.match(value)

    if res is None:
        if _debug:
            print("no match")
    else:

        ret = {'min': None, 'sec': None}

        ret_attributes.add('truncated_time')
        groupdict = res.groupdict()

        if groupdict['number1'] not in [None, '-']:
            ret_attributes.add('min')
            ret['min'] = int(groupdict['number1'])

        if groupdict['number2'] not in [None]:
            ret_attributes.add('sec')
            ret['sec'] = int(groupdict['number2'])

        if groupdict['sep'] == ':':
            ret_attributes.add(':')

    return ret, ret_attributes


def date_ordinal_str_to_date(value):

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    _debug = False

    ret = None
    ret_attributes = set()

    res = DATE_ORDINAL_EXPRESSION_C.match(value)

    if res is None:
        if _debug:
            print("no match")
    else:

        groupdict = res.groupdict()
        if _debug:
            print(repr(groupdict))
        separator_error = False
        separator = groupdict['sep']

        for i in ['year', 'day']:
            if groupdict[i] is None:
                groupdict[i] = 1
            else:
                ret_attributes.add(i)

        if separator_error:
            if _debug:
                print("separator error")
        else:

            if separator == '-':
                ret_attributes.add(separator)

            ret_attributes.add('ordinal')

            ret = datetime.date(
                int(groupdict['year']),
                1,
                1
                ) + datetime.timedelta(days=int(groupdict['day']))

            _date_attrs_check(ret_attributes)

    return ret, ret_attributes


def date_week_str_to_date(value):

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    _debug = False

    ret = None
    ret_attributes = set()

    res = DATE_WEEK_EXPRESSION_C.match(value)

    if res is None:
        if _debug:
            print("no match")
    else:

        groupdict = res.groupdict()
        if _debug:
            print(repr(groupdict))
        separator_error = False
        separator = groupdict['sep1']

        if (groupdict['day'] is not None
                and groupdict['sep2'] != separator):
            separator_error = True

        for i in ['week', 'day']:
            if groupdict[i] is None:
                groupdict[i] = '01'
            else:
                ret_attributes.add(i)

        for i in ['year']:
            if groupdict[i] is None:
                groupdict[i] = '0001'
            else:
                ret_attributes.add(i)

        if separator_error:
            if _debug:
                print("separator error")
        else:

            ret_attributes.add('week')

            if separator == '-':
                ret_attributes.add(separator)

            ret = datetime.date(
                int(groupdict['year']),
                1,
                1
                ) + datetime.timedelta(
                    days=int(groupdict['day']),
                    weeks=int(groupdict['week'])
                    )
            _date_attrs_check(ret_attributes)

    return ret, ret_attributes


def date_normal_str_to_date(value):

    if not isinstance(value, str):
        raise TypeError("`value' must be inst of str")

    _debug = False

    ret = None
    ret_attributes = set()

    res = DATE_EXPRESSION_C.match(value)

    if res is None:
        if _debug:
            print("no match")
    else:

        groupdict = res.groupdict()
        if _debug:
            print(repr(groupdict))
        separator_error = False
        separator = groupdict['sep1']

        if (groupdict['day'] is not None
                and groupdict['sep2'] != separator):
            separator_error = True

        if len(groupdict['year']) == 2 and groupdict['not_year'] is None:
            groupdict['year'] += '00'
            ret_attributes.add('century')
        else:
            ret_attributes.add('year')

        for i in ['month', 'day']:
            if groupdict[i] is None:
                groupdict[i] = '01'
            else:
                ret_attributes.add(i)

        if separator_error:
            if _debug:
                print("separator error")
        else:

            if separator == '-':
                ret_attributes.add(separator)

            ret = datetime.date(
                int(groupdict['year']),
                int(groupdict['month']),
                int(groupdict['day'])
                )

            _date_attrs_check(ret_attributes)

    return ret, ret_attributes


def dict_to_truncated_date_str(value, attr):

    if not isinstance(value, dict):
        raise TypeError("`value' must be inst of dict")

    day = ''
    if 'day' in attr and value['day'] is not None:
        day = '{:02d}'.format(int(value['day']))

    month = ''
    if 'month' in attr and value['month'] is not None or value['month'] == '-':
        month = '{:02d}'.format(int(value['month']))

    if not 'month' in attr:
        month = '-'

    sep = ''
    if '-' in attr and day != '' and not month in ['', '-']:
        sep = '-'

    ret = \
        '--{month}{sep}{day}'.format(
            month=month,
            sep=sep,
            day=day
            )

    return ret


def dict_to_truncated_time_str(value, attr):

    if not isinstance(value, dict):
        raise TypeError("`value' must be inst of dict")

    sec = ''
    if 'sec' in attr and value['sec'] is not None:
        sec = '{:02d}'.format(int(value['sec']))

    minu = ''
    if 'min' in attr and value['min'] is not None or value['min'] == '-':
        minu = '{:02d}'.format(int(value['min']))

    if not 'min' in attr:
        minu = '-'

    sep = ''
    if ':' in attr and sec != '' and not minu in ['', '-']:
        sep = ':'

    ret = \
        '-{minu}{sep}{sec}'.format(
            minu=minu,
            sep=sep,
            sec=sec
            )

    return ret


def gen_tz(h, m, plus=True):

    td = datetime.timedelta(hours=h, minutes=m)
    if not plus:
        td = -td

    tz = datetime.timezone(td)

    return tz


def format_tz(value, sep=True, minu=True, zed=False):

    if value is not None and not isinstance(value, datetime.timezone):
        raise TypeError("`value' must be None or inst of datetime.timezone")

    ret = None

    if value is None:
        ret = ''
    else:
        a = value.utcoffset(None)

        if (value == datetime.timezone.utc or a.seconds == 0) and zed == True:
            # TODO: looks like here is needed more checks
            ret = 'Z'
        else:

            sign = '+'

            if a < datetime.timedelta():
                sign = '-'
                a = -a

            separator = ''
            if sep and minu:
                separator = ':'

            hours = int(a.seconds / 60 / 60)

            minutes = ''
            if minu:
                minutes = int((a.seconds - (hours * 60 * 60)) / 60 / 60)

            ret = '{sign}{hour}{sep}{minute}'.format(
                sign=sign,
                hour='{:02d}'.format(hours),
                sep=separator,
                minute='{:02d}'.format(minutes)
                )

    return ret


def _test(variants, callab, callab2):

    for i in variants:

        print(i)

        res = callab(i)

        if res[0] is None:
            print("'{}' not matches".format(i))
        else:
            print("forward result {}".format(repr(res)))

            res = callab2(res[0], res[1])

            if res is None:
                print("'{}' not matches".format(i))
            else:
                print("reverse result\n{}".format(res))

        print()

    return


def test_date():

    variants = [
        '19850412',
        '1985-04-12',
        '1985-04',
        '1985',
        '19',
        '1985102',
        '-1985102',
        '1985-102',
        '1985W155',
        '1985-W15-5',
        '1985W15',
        '1985-W15',
        '2007-01-25',
        '--0412',
        '--04-12',
        '--04',
        '---12'
        ]

    _test(variants, str_to_date, date_to_str)

    return


def test_time():

    variants = [
        '232050',
        '23:20:50',
        '2320',
        '23:20',
        '23',
        '02:56:15Z',
        '21:56:15-05:00',
        '02:56:15,66Z',
        '21:56:15.99-05:00',
        '215615.99-0500',
        '02:56:15,66Z',
        '02:56,66Z',
        '02,66Z',
        '232050,5',
        '23:20:50,5',
        '2320,8',
        '23:20,8',
        '23,3',
        '12:00:00Z',
        '-2050',
        '-20:50',
        '-20',
        '--50'
        ]

    _test(variants, str_to_time, time_to_str)

    return


def test_datetime():

    variants = [
        '2007-01-25T12:00:00Z'
        ]

    _test(variants, str_to_datetime, datetime_to_str)

    return
